from sila.framework.data_types import Any as AnyType
from sila.framework.data_types import (
    Binary,
    Boolean,
    Constrained,
    DataType,
    Date,
    Duration,
    Integer,
    List,
    Real,
    String,
    Structure,
    Time,
    Timestamp,
    Timezone,
    Void,
)

from .convert_data_type import Any, from_sila, to_sila
from .custom_data_type import CustomDataType
from .infer_data_type import infer

__all__ = [
    "Any",
    "AnyType",
    "Binary",
    "Boolean",
    "Constrained",
    "CustomDataType",
    "DataType",
    "Date",
    "Duration",
    "Integer",
    "List",
    "Real",
    "String",
    "Structure",
    "Time",
    "Timestamp",
    "Timezone",
    "Void",
    "from_sila",
    "infer",
    "to_sila",
]



import asyncio
import re
import typing
import unittest.mock
import weakref

import pytest

from unitelabs.cdk.subscriptions import Subject, Subscription
from unitelabs.cdk.subscriptions.subject import _DEFAULT_VALUE

ONE_OP_TIME = 0.01


def first_pipe(x: str) -> str:
    return x.upper()


def second_pipe(x: str) -> dict[str, str]:
    return {"value": x}


async def redundant_update(subject: Subject[str]) -> None:
    """Update the subject 10x, once per operation time, with redundant updates after the first iteration."""
    for x in range(1, 11):
        if x > 1:
            subject.update(f"update {x - 1}")  # redundant update
        subject.update(f"update {x}")
        await asyncio.sleep(ONE_OP_TIME)


class TestDefaults:
    async def test_should_set_default_value_as_current(self):
        subject = Subject[str]()
        assert subject.current is _DEFAULT_VALUE

    async def test_should_set_default_maxsize(self):
        subject = Subject[str]()
        sub = subject.subscribe()
        assert sub.maxsize == 0

    async def test_should_allow_optional_value(self):
        subject = Subject[typing.Optional[str]](maxsize=10)
        subject.update(None)
        assert subject.current is None


class TestInit:
    async def test_should_type_check_pipe(self):
        # Requires visual check
        def invalid_pipe(x: int) -> str:
            return "value"

        Subject[str](maxsize=10, pipe=invalid_pipe)

    async def test_should_allow_pipe_function(self):
        def pipe_func(x: str) -> str:
            return x.lower()

        subject = Subject[str](maxsize=10, pipe=pipe_func)
        assert subject._pipe == pipe_func
        subject.update("TEST")
        assert subject.current == "test"


class TestUpdate:
    async def test_should_update_current_value(self):
        subject = Subject[str](maxsize=10)
        subject.update("new value")
        assert subject.current == "new value"

    async def test_should_notify_even_if_value_is_current(self):
        subject = Subject[str](maxsize=10)
        subject.update("new value")
        subject.notify = unittest.mock.Mock(wraps=subject.notify)
        assert isinstance(subject.current, str)
        subject.update(subject.current)
        subject.notify.assert_called_once()


class TestSubscribe:
    async def test_should_add_subscription(self):
        subject = Subject[str](maxsize=10)
        sub = subject.subscribe()
        assert sub in subject.subscribers

    async def test_should_not_set_current_value_on_subscription(self, create_task):
        subject = Subject[str](maxsize=10)
        FIRST_UPDATE = "first"
        subject.update(FIRST_UPDATE)
        assert subject.current == FIRST_UPDATE

        sub = subject.subscribe()
        assert subject.current == FIRST_UPDATE

        update_task = next(create_task(redundant_update(subject)))
        assert await sub.get() == "update 1"
        assert subject.current == "update 1"

        update_task.cancel()
        await asyncio.sleep(0.01)
        assert update_task.done()
        subject.unsubscribe(sub)


class TestOnSubscribe:
    async def test_should_trigger_on_subscription(self):
        subject = Subject[str](maxsize=10)
        subject.on_subscribe = unittest.mock.Mock(wraps=subject.on_subscribe)

        subject.subscribe()

        subject.on_subscribe.assert_called_once_with()

    async def test_should_trigger_on_child_subscription(self):
        subject = Subject[str](maxsize=10)
        subject.on_subscribe = unittest.mock.Mock(wraps=subject.on_subscribe)
        child_subject = subject.pipe(first_pipe).pipe(second_pipe)

        child_subject.subscribe()

        subject.on_subscribe.assert_called_once_with()

    async def test_should_trigger_only_on_first_subscription(self):
        subject = Subject[str](maxsize=10)
        subject.on_subscribe = unittest.mock.Mock(wraps=subject.on_subscribe)
        child_subject = subject.pipe(first_pipe).pipe(second_pipe)

        subject.subscribe()
        subject.on_subscribe.assert_called_once_with()
        subject.on_subscribe.reset_mock()

        subject.subscribe()
        child_subject.subscribe()
        subject.on_subscribe.assert_not_called()

    async def test_should_trigger_only_on_first_child_subscription(self):
        subject = Subject[str](maxsize=10)
        subject.on_subscribe = unittest.mock.Mock(wraps=subject.on_subscribe)
        child_subject = subject.pipe(first_pipe).pipe(second_pipe)

        child_subject.subscribe()
        subject.on_subscribe.assert_called_once_with()
        subject.on_subscribe.reset_mock()

        subject.subscribe()
        child_subject.subscribe()
        subject.on_subscribe.assert_not_called()


class TestOnUnsubscribe:
    async def test_should_trigger_on_unsubscription(self):
        subject = Subject[str](maxsize=10)
        subject.on_unsubscribe = unittest.mock.Mock(wraps=subject.on_unsubscribe)

        sub = subject.subscribe()
        subject.unsubscribe(sub)

        subject.on_unsubscribe.assert_called_once_with()

    async def test_should_trigger_on_child_unsubscription(self):
        subject = Subject[str](maxsize=10)
        subject.on_unsubscribe = unittest.mock.Mock(wraps=subject.on_unsubscribe)
        child_subject = subject.pipe(first_pipe).pipe(second_pipe)

        sub = child_subject.subscribe()
        subject.unsubscribe(sub)

        subject.on_unsubscribe.assert_called_once_with()

    async def test_should_trigger_only_on_last_unsubscription(self):
        subject = Subject[str](maxsize=10)
        subject.on_unsubscribe = unittest.mock.Mock(wraps=subject.on_unsubscribe)
        child_subject = subject.pipe(first_pipe).pipe(second_pipe)

        sub1 = subject.subscribe()
        sub2 = child_subject.subscribe()
        sub3 = subject.subscribe()

        subject.unsubscribe(sub1)
        child_subject.unsubscribe(sub2)
        subject.on_unsubscribe.assert_not_called()

        subject.unsubscribe(sub3)
        subject.on_unsubscribe.assert_called_once_with()

    async def test_should_trigger_only_on_last_child_unsubscription(self):
        subject = Subject[str](maxsize=10)
        subject.on_unsubscribe = unittest.mock.Mock(wraps=subject.on_unsubscribe)
        child_subject = subject.pipe(first_pipe).pipe(second_pipe)

        sub1 = subject.subscribe()
        sub2 = child_subject.subscribe()
        sub3 = child_subject.subscribe()

        subject.unsubscribe(sub1)
        subject.unsubscribe(sub2)
        subject.on_unsubscribe.assert_not_called()

        subject.unsubscribe(sub3)
        subject.on_unsubscribe.assert_called_once_with()


class TestUnsubscribe:
    async def test_should_remove_subscription(self):
        subject = Subject[str](maxsize=10)
        sub = subject.subscribe()
        assert sub in subject.subscribers
        subject.unsubscribe(sub)
        assert sub not in subject.subscribers

    async def test_should_raise_value_error_on_unknown_subscription(self):
        subject = Subject[str](maxsize=10)
        with pytest.raises(ValueError, match="Subscription not found in subscribers or children."):
            subject.unsubscribe(Subscription(10, parent=unittest.mock.Mock()))

    async def test_should_raise_value_error_on_twice_removed(self):
        subject = Subject[str](maxsize=10)
        sub = subject.subscribe()
        subject.unsubscribe(sub)
        with pytest.raises(ValueError, match="Subscription not found in subscribers or children."):
            subject.unsubscribe(sub)

    async def test_should_cancel_subscription(self):
        subject = Subject[str](maxsize=10)
        sub = subject.subscribe()
        sub.cancel = unittest.mock.Mock(wraps=sub.cancel)

        subject.unsubscribe(sub)

        sub.cancel.assert_called_once()
        assert sub not in subject.subscribers

        subject.update("value")
        assert not sub.size  # should not receive updates after being unsubscribed

    async def test_should_allow_unsubscribe_from_direct_parent_subject(self):
        subject = Subject[str](maxsize=10)
        child_subject = subject.pipe(first_pipe).pipe(second_pipe)

        sub = child_subject.subscribe()
        assert sub not in subject.subscribers
        assert sub in child_subject.subscribers

        subject.unsubscribe(sub)
        assert sub not in child_subject.subscribers

        with pytest.raises(TimeoutError):
            await sub.get(timeout=0.01)

    async def test_should_allow_unsubscribe_from_ancestral_parent_subject(self):
        subject = Subject[str](maxsize=10)
        child_subject = subject.pipe(first_pipe).pipe(second_pipe)

        sub = child_subject.subscribe()
        assert sub not in subject.subscribers
        assert sub in child_subject.subscribers

        subject.unsubscribe(sub)
        assert sub not in child_subject.subscribers

        with pytest.raises(TimeoutError):
            await sub.get(timeout=0.01)

    async def test_should_not_allow_unsubscribe_from_sibling_subject(self):
        subject = Subject[str](maxsize=10)
        child_subject = subject.pipe(first_pipe).pipe(second_pipe)
        other_subject = subject.pipe(second_pipe)

        sub = child_subject.subscribe()
        assert sub not in subject.subscribers
        assert sub in child_subject.subscribers

        with pytest.raises(ValueError, match="Subscription not found in subscribers or children."):
            other_subject.unsubscribe(sub)

    async def test_should_prune_temporary_child_subjects(self):
        subject = Subject[str](maxsize=10)
        child_subject = subject.pipe(first_pipe, temporary=True)
        grandchild_subject = child_subject.pipe(second_pipe, temporary=True)

        sub = grandchild_subject.subscribe()
        assert subject._children
        assert subject._children[0]._children

        subject.unsubscribe(sub)
        assert not subject._children

    async def test_should_not_prune_child_subjects_with_subscriptions(self):
        subject = Subject[str](maxsize=10)
        child_subject = subject.pipe(first_pipe)
        temp_child_subject = child_subject.pipe(second_pipe, temporary=True)

        child_subject.subscribe()
        temp_grandchild_sub = temp_child_subject.subscribe()

        assert child_subject in subject._children
        assert temp_child_subject in child_subject._children

        subject.unsubscribe(temp_grandchild_sub)
        assert child_subject in subject._children
        assert not child_subject._children

    async def test_should_not_prune_parent_subjects_with_subscriptions(self):
        subject = Subject[str](maxsize=10)
        temp_child_subject = subject.pipe(first_pipe, temporary=True)
        assert temp_child_subject in subject._children

        with pytest.raises(
            RuntimeError,
            match=(
                re.escape(
                    "Cannot create a non-temporary `Subject` from a temporary `Subject`, use pipe() with `temporary=True` "
                    "or adjust the current `Subject` to be non-temporary."
                )
            ),
        ):
            temp_child_subject.pipe(second_pipe)


class TestPipes:
    async def test_should_type_check_pipes(self):
        # Requires visual check
        def invalid_pipe(x: int) -> str:
            return "value"

        subject = Subject[str](maxsize=10)
        subject.pipe(invalid_pipe)

    async def test_should_create_new_subject(self):
        subject = Subject[str](maxsize=10)
        piped = subject.pipe(first_pipe).pipe(second_pipe)

        assert isinstance(piped, Subject)
        assert piped != subject

    async def test_should_not_set_pipes_on_current_subject(self):
        subject = Subject[str](maxsize=10)
        start_pipe = subject._pipe

        upper_subject = subject.pipe(first_pipe)
        assert subject._pipe == start_pipe
        assert upper_subject._pipe == first_pipe

        to_dict_subject = subject.pipe(second_pipe)
        assert subject._pipe == start_pipe
        assert to_dict_subject._pipe == second_pipe

        # test that pipes are applied correctly
        upper_subscription = upper_subject.subscribe()
        to_dict_subscription = to_dict_subject.subscribe()

        subject.update("value")
        assert await upper_subscription.get() == "VALUE"
        assert await to_dict_subscription.get() == {"value": "value"}

    async def test_should_track_child_subscription(self):
        subject = Subject[str](maxsize=10)
        piped = subject.pipe(first_pipe).pipe(second_pipe)

        sub = piped.subscribe()

        assert subject._children[0]._pipe == first_pipe
        assert subject._children[0]._children[0]._pipe == second_pipe

        assert sub in piped.subscribers
        assert piped._parent._parent == weakref.proxy(subject)

        # cleanup
        piped.unsubscribe(sub)

    async def test_should_update_value_in_all_child_subjects(self):
        subject = Subject[str](maxsize=10)
        sub = subject.subscribe()
        sub.update = unittest.mock.Mock(wraps=sub.update)

        piped = subject.pipe(first_pipe).pipe(second_pipe)
        piped.update = unittest.mock.Mock(wraps=piped.update)

        piped_sub = piped.subscribe()
        piped_sub.update = unittest.mock.Mock(wraps=piped_sub.update)

        subject.update("value")

        sub.update.assert_called_once_with("value")
        piped.update.assert_called_once_with("VALUE")
        piped_sub.update.assert_called_once_with({"value": "VALUE"})
        assert await sub.get() == "value"
        assert await piped_sub.get() == {"value": "VALUE"}

        # cleanup
        for s in [sub, piped_sub]:
            subject.unsubscribe(s)


class TestFilter:
    async def test_should_allow_filtering(self):
        subject = Subject[str](maxsize=10)
        subscription = subject.subscribe()
        filtered = subject.filter(lambda x: x and "filter" not in x)
        filtered.update = unittest.mock.Mock(wraps=filtered.update)
        filtered_subscription = filtered.subscribe()
        filtered_subscription.update = unittest.mock.Mock(wraps=filtered_subscription.update)

        subject.update("filter this")
        filtered.update.assert_called_once_with("filter this")
        filtered_subscription.update.assert_not_called()
        assert filtered_subscription.empty()
        assert await subscription.get() == "filter this"

    async def test_should_pass_filtering(self):
        subject = Subject[str](maxsize=10)
        subscription = subject.subscribe()
        filtered = subject.filter(lambda x: x and "filter" not in x)
        filtered.update = unittest.mock.Mock(wraps=filtered.update)
        filtered_subscription = filtered.subscribe()
        filtered_subscription.update = unittest.mock.Mock(wraps=filtered_subscription.update)

        subject.update("pass")
        filtered.update.assert_called_once_with("pass")
        filtered_subscription.update.assert_called_once_with("pass")
        assert await filtered_subscription.get() == "pass"
        assert await subscription.get() == "pass"

    async def test_should_prune_from_parent_if_temporary(self):
        subject = Subject[int](maxsize=10)
        filtered = subject.filter(lambda x: x > 5, temporary=True)
        assert filtered in subject._children

        filtered_subscription = filtered.subscribe()
        assert filtered_subscription._parent._is_temporary

        subject.unsubscribe(filtered_subscription)
        assert filtered not in subject._children


class TestContextManager:
    async def test_should_subscribe_on_enter(self):
        subject = Subject[str](maxsize=10)

        with subject as subscription:
            assert subscription in subject.subscribers

    async def test_should_unsubscribe_on_exit(self):
        subject = Subject[str](maxsize=10)

        with subject:
            pass

        assert not subject.subscribers

    async def test_should_raise_on_nested_contexts(self):
        subject = Subject[str](maxsize=10)

        with subject, pytest.raises(RuntimeError), subject:
            pass

        assert not subject.subscribers

    async def test_should_reset__context(self):
        subject = Subject[int](maxsize=10)
        with subject as subscription:
            assert subscription in subject.subscribers
        with subject as subscription:
            assert subscription in subject.subscribers

    async def test_should_allow_and_tear_down_temporary_pipe(self):
        subject = Subject[int](maxsize=10)
        with subject.pipe(lambda x: x * 2, temporary=True) as subscription:
            for i in range(10):
                subject.update(i)
                value = await subscription.get()
                assert value == i * 2

        assert not subject.subscribers
        assert not subject._children


class TestSubscription_Get:
    async def test_should_get_new_value(self):
        subject = Subject[str](maxsize=10)
        subscription: Subscription[str] = subject.subscribe()
        subscription.update = unittest.mock.Mock(wraps=subscription.update)

        for i in range(10):
            msg = f"update {i}"
            subject.update(msg)
            subscription.update.assert_called_with(msg)
            assert await subscription.get() == msg

        # cleanup
        subject.unsubscribe(subscription)

    async def test_should_timeout_if_nothing_queued(self):
        subject = Subject[str](maxsize=10)
        subscription: Subscription[str] = subject.subscribe()
        with pytest.raises(TimeoutError):
            await subscription.get(timeout=0.05)

        # cleanup
        subject.unsubscribe(subscription)

    async def test_should_timeout_if_nothing_matches_predicate(self, create_task):
        subject = Subject[str](maxsize=10)
        update_task = next(create_task(redundant_update(subject)))
        subscription: Subscription[str] = subject.subscribe()

        with pytest.raises(TimeoutError):
            await subscription.get(lambda x: "value" in x, timeout=0.05)

        # cleanup
        update_task.cancel()
        await asyncio.sleep(0.01)
        assert update_task.done()
        subject.unsubscribe(subscription)

    async def test_should_not_set_subject_current_value_on_new_subscriptions(self):
        subject = Subject[Exception](maxsize=10)
        subscription = subject.subscribe()

        exception = ValueError("Test exception")
        subject.update(exception)

        expected_exception = await subscription.get(lambda x: isinstance(x, ValueError))
        assert expected_exception == exception
        assert subject._value == exception

        new_subscription = subject.subscribe()
        with pytest.raises(TimeoutError):
            await new_subscription.get(lambda x: isinstance(x, ValueError), timeout=0.05)
        assert subject._value == exception

        # cleanup
        for s in [subscription, new_subscription]:
            subject.unsubscribe(s)



# ruff: noqa: D401, D415

import asyncio

from unitelabs.cdk import sila, subscriptions


class ObservablePropertyTest(sila.Feature):
    """This is a test feature to test observable properties."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

        self.alternating = False
        self.alternating_publisher = subscriptions.Publisher(self._update_alternating, interval=1)

        self.value: int = 1
        self.value_event = asyncio.Event()

    async def _update_alternating(self) -> bool:
        self.alternating = not self.alternating

        return self.alternating

    @sila.ObservableProperty()
    async def subscribe_fixed_value(self) -> sila.Stream[int]:
        """Always returns 42 and never changes."""

        yield 42

    @sila.ObservableProperty()
    async def subscribe_alternating(self) -> sila.Stream[bool]:
        """Switches every second between true and false"""

        return self.alternating_publisher.subscribe()

    @sila.ObservableProperty()
    async def subscribe_editable(self) -> sila.Stream[int]:
        """Can be set through SetValue command"""

        while True:
            self.value_event.clear()
            yield self.value
            await self.value_event.wait()

    @sila.UnobservableCommand()
    def set_value(self, value: int) -> None:
        """
        Changes the value of Editable

        .. parameter:: The new value
        """

        self.value = value
        self.value_event.set()



import dataclasses
import inspect
import warnings
import weakref

import typing_extensions as typing

from sila.server import DataType, Element, Feature, Structure

from .. import utils


@dataclasses.dataclass
class Dataclass:
    """Base class for dataclass based SiLA annotations."""

    _identifier: typing.ClassVar[str] = ""
    _name: typing.ClassVar[str] = ""
    _description: typing.ClassVar[str] = ""

    def __init_subclass__(
        cls,
        /,
        *,
        identifier: typing.Optional[str] = None,
        display_name: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
    ) -> None:
        super().__init_subclass__()

        if display_name is not None:
            msg = "Using `display_name` is deprecated, use `name` instead."
            warnings.warn(msg, stacklevel=2)
            name = display_name

        cls._name = name or utils.to_display_name(cls.__name__)
        cls._identifier = identifier or cls._name.replace(" ", "")
        cls._description = utils.parse_docs(inspect.getdoc(cls)).get("default", "")
        cls._feature: typing.Optional["Feature"] = None

    @classmethod
    def attach(cls, feature: "Feature") -> None:
        """
        Create and attach a dataclass to the `feature`.

        Args:
          feature: The `Feature` to which the dataclass will be attached.
        """

        cls._feature = weakref.proxy(feature)

    @classmethod
    def _infer_data_type(cls, feature: Feature) -> type[DataType]:
        docs = utils.parse_docs(inspect.getdoc(cls))

        from ..data_types.infer_data_type import infer

        fields = dataclasses.fields(cls)
        data_type: type[DataType]

        if len(fields) == 0:
            msg = (
                f"Could not detect any fields on '{cls._identifier}'. "
                "Did you forget to annotated your data type definition with `@dataclasses.dataclass`?"
            )
            raise ValueError(msg)
        if len(fields) == 1 and utils.humanize(fields[0].name).replace(" ", "") == cls._identifier:
            data_type = infer(fields[0].type, feature)
        else:
            elements: dict[str, Element] = {}
            for index, field in enumerate(fields):
                field_display_name = utils.humanize(field.name)
                field_identifier = field_display_name.replace(" ", "")
                parameter = docs.get("parameter", [])
                elements[field.name] = Element(
                    identifier=field_identifier,
                    display_name=field_display_name,
                    description=parameter[index].get("default", "") if len(parameter) > index else "",
                    data_type=infer(field.type, feature),
                )
            data_type = Structure.create(elements=elements, name=cls.__name__)

        return data_type



from .config import (
    Config,
    ConfigurationError,
    Field,
    UnsupportedConfigFiletype,
    delayed_default,
    read_config_file,
    validate_config,
    validate_field,
)
from .connector_base_config import CloudServerConfig, ConnectorBaseConfig, SiLAServerConfig, get_connector_config
from .schema import InvalidSchemaFieldError, describe

__all__ = [
    "CloudServerConfig",
    "Config",
    "ConfigurationError",
    "ConnectorBaseConfig",
    "Field",
    "InvalidSchemaFieldError",
    "SiLAServerConfig",
    "UnsupportedConfigFiletype",
    "delayed_default",
    "describe",
    "get_connector_config",
    "read_config_file",
    "validate_config",
    "validate_field",
]



# ruff: noqa: D401, E501

from unitelabs.cdk import sila


class BasicDataTypesTest(sila.Feature):
    """Provides commands and properties to set or respectively get all SiLA Basic Data Types via command parameters or property responses respectively."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    # Data type String

    @sila.UnobservableCommand()
    @sila.Response(name="Received Value")
    def echo_string_value(self, string_value: str) -> str:
        """
        Receives a String value and returns the String value that has been received.

        .. parameter:: The String value to be returned.
        .. return:: The String value that has been received.
        """

        return string_value

    @sila.UnobservableProperty()
    def get_string_value(self) -> str:
        """Returns the String value 'SiLA2_Test_String_Value'."""

        return "SiLA2_Test_String_Value"

    # Data type Integer

    @sila.UnobservableCommand()
    @sila.Response(name="Received Value")
    def echo_integer_value(self, integer_value: int) -> int:
        """
        Receives an Integer value and returns the Integer value that has been received.

        .. parameter:: The Integer value to be returned.
        .. return:: The Integer value that has been received.
        """

        return integer_value

    @sila.UnobservableProperty()
    def get_integer_value(self) -> int:
        """Returns the Integer value 5124."""

        return 5124

    # Data type Real

    @sila.UnobservableCommand()
    @sila.Response(name="Received Value")
    def echo_real_value(self, real_value: float) -> float:
        """
        Receives a Real value and returns the Real value that has been received.

        .. parameter:: The Real value to be returned.
        .. return:: The Real value that has been received.
        """

        return real_value

    @sila.UnobservableProperty()
    def get_real_value(self) -> float:
        """Returns the Real value 3.1415926."""

        return 3.1415926

    # Data type Boolean

    @sila.UnobservableCommand()
    @sila.Response(name="Received Value")
    def echo_boolean_value(self, boolean_value: bool) -> bool:
        """
        Receives a Boolean value and returns the Boolean value that has been received.

        .. parameter:: The Boolean value to be returned.
        .. return:: The Boolean value that has been received.
        """

        return boolean_value

    @sila.UnobservableProperty()
    def get_boolean_value(self) -> bool:
        """Returns the Boolean value true."""

        return True

    # Data type Date

    @sila.UnobservableCommand()
    @sila.Response(name="Received Value")
    def echo_date_value(self, date_value: sila.datetime.date) -> sila.datetime.date:
        """
        Receives a Date value and returns the Date value that has been received.

        .. parameter:: The Date value to be returned.
        .. return:: The Date value that has been received.
        """

        return date_value

    @sila.UnobservableProperty()
    def get_date_value(self) -> sila.datetime.date:
        """Returns the Date value 05.08.2022 respective 08/05/2018, timezone +2."""

        return sila.datetime.date(
            year=2022, month=8, day=5, tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2))
        )

    # Data type Time

    @sila.UnobservableCommand()
    @sila.Response(name="Received Value")
    def echo_time_value(self, time_value: sila.datetime.time) -> sila.datetime.time:
        """
        Receives a Time value and returns the Time value that has been received.

        .. parameter:: The Time value to be returned.
        .. return:: The Time value that has been received.
        """

        return time_value

    @sila.UnobservableProperty()
    def get_time_value(self) -> sila.datetime.time:
        """Returns the Time value 12:34:56.789, timezone +2."""

        return sila.datetime.time(
            hour=12,
            minute=34,
            second=56,
            microsecond=789000,
            tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
        )

    # Data type Timestamp

    @sila.UnobservableCommand()
    @sila.Response(name="Received Value")
    def echo_timestamp_value(self, timestamp_value: sila.datetime.datetime) -> sila.datetime.datetime:
        """
        Receives a Timestamp value and returns a message containing the Timestamp value that has been received.

        .. parameter:: The Timestamp value to be returned.
        .. return:: The Timestamp value that has been received.
        """

        return timestamp_value

    @sila.UnobservableProperty()
    def get_timestamp_value(self) -> sila.datetime.datetime:
        """Returns the Timestamp value 2022-08-05 12:34:56.789, timezone +2."""

        return sila.datetime.datetime(
            year=2022,
            month=8,
            day=5,
            hour=12,
            minute=34,
            second=56,
            microsecond=789000,
            tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
        )



import asyncio
import collections.abc
import inspect
import re
import time
import typing

import packaging.version


def humanize(value: typing.Optional[str] = None, sep: str = "_") -> str:
    """Convert `value` to a human-readable string."""

    return " ".join(x.capitalize() for x in (value or "").split(sep))


def parse_docs(docs: typing.Optional[str] = None) -> dict:
    """Parse documentation strings."""
    docs = docs or ""
    directives = re.split(r"^\.\. *([^:]+):: *", docs, flags=re.MULTILINE)
    result = {"default": inspect.cleandoc(directives.pop(0)).replace("\n", " ")}

    for i in range(0, len(directives), 2):
        key = directives[i]

        params = re.split(r"^ *:([^:]+): *", directives[i + 1], flags=re.MULTILINE)
        item = {"default": inspect.cleandoc(params.pop(0)).replace("\n", " ")}

        par = {params[i]: params[i + 1] for i in range(0, len(params), 2)}
        for param_k, param_v in par.items():
            item[param_k] = inspect.cleandoc(param_v).replace("\n", " ")

        result[key] = result.get(key, [])
        result[key].append(item)

    return result


def parse_version(version: str) -> str:
    """Parse a PEP 440-compatible version into a valid SiLA 2 version."""

    try:
        v = packaging.version.parse(version)
        parts = [str(v.major), str(v.minor)]
        if v.micro != 0:
            parts.append(str(v.micro))
        sila_version = ".".join(parts)

        suffix_parts = []
        if v.pre:
            suffix_parts.append(f"{v.pre[0]}{v.pre[1]}")
        if v.post:
            suffix_parts.append(f"post{v.post}")
        if v.dev:
            suffix_parts.append(f"dev{v.dev}")
        if v.local:
            suffix_parts.append(v.local.replace(".", "_"))

        if suffix_parts:
            sila_version += "_" + "_".join(suffix_parts)

        return sila_version
    except packaging.version.InvalidVersion:
        msg = f"Invalid version format: '{version}'."
        raise ValueError(msg) from None


def to_display_name(value: str) -> str:
    """Convert `value` to a display name."""
    value = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1 \2", value)
    value = re.sub(r"([a-z\d])([A-Z])", r"\1 \2", value)
    return value.replace("-", " ")


def set_interval(function: collections.abc.Callable, delay: float = 1) -> asyncio.Task:
    """Repeatedly call a function or execute a codesnippet, with a fixed time delay between each call."""
    delay_ns = delay * 10**9
    timer = time.perf_counter_ns()

    async def interval(timer: float) -> None:
        while True:
            response = function()
            if inspect.isawaitable(response):
                await response

            timer += delay_ns
            await asyncio.sleep((timer - time.perf_counter_ns()) / 10**9)

    return asyncio.create_task(interval(timer))


def clear_interval(interval: asyncio.Task) -> None:
    """Cancel a timed, repeating action which was previously established by a call to set_interval()."""
    interval.cancel()


__all__ = ["clear_interval", "humanize", "parse_docs", "set_interval", "to_display_name"]



import click
import dotenv

from .dev import dev
from .start import start
from .sync import sync

CONTEXT_SETTINGS = dict(show_default=True)


@click.group(context_settings=CONTEXT_SETTINGS)
def connector() -> click.Group:
    """Connector commands."""
    dotenv.load_dotenv()


connector.add_command(dev)
connector.add_command(start)
connector.add_command(sync)

if __name__ == "__main__":
    connector()



# ruff: noqa: D205

import asyncio
import datetime

import typing_extensions as typing

from unitelabs.cdk import sila


class ObservableCommandTest(sila.Feature):
    """
    This is a test feature to test observable commands.
    It specifies various observable commands and returns defined answers to validate against.
    """

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    @sila.ObservableCommand()
    @sila.IntermediateResponse(name="Current Iteration")
    @sila.Response(name="Iteration Response")
    async def count(
        self,
        n: int,
        delay: typing.Annotated[
            float, sila.constraints.Unit(label="s", components=[sila.constraints.Unit.Component("Second")])
        ],
        *,
        status: sila.Status,
        intermediate: sila.Intermediate[int],
    ) -> int:
        """
        Count from 0 to N-1 and return the current number as intermediate response.

        .. parameter:: Number to count to
        .. parameter:: The delay for each iteration
        .. yield:: The current number, from 0 to N-1 (excluded).
        .. return:: The last number (N-1)
        """

        for i in range(n):
            status.update(
                progress=i / (n - 1),
                remaining_time=datetime.timedelta(seconds=delay * (n - i - 1)),
            )
            intermediate.send(i)

            await asyncio.sleep(delay)

        return n - 1

    @sila.ObservableCommand()
    @sila.Response(name="Received Value")
    async def echo_value_after_delay(
        self,
        value: int,
        delay: typing.Annotated[
            float, sila.constraints.Unit(label="s", components=[sila.constraints.Unit.Component("Second")])
        ],
        *,
        status: sila.Status,
    ) -> int:
        """
        Echo the given value after the specified delay. The command state must be "waiting" until the delay has passed.

        .. parameter:: The value to echo
        .. parameter:: The delay before the command execution starts
        .. return:: The Received Value
        """

        seconds, rest = divmod(delay, 1)
        for i in range(int(seconds)):
            await asyncio.sleep(1)
            status.update(progress=i / delay, remaining_time=datetime.timedelta(seconds=delay - i))

        await asyncio.sleep(rest)
        return value



import collections.abc
import dataclasses
import inspect
import itertools
import warnings

from sila.server import Element, Feature

from ..data_types import infer


@dataclasses.dataclass
class Decorator:
    """
    A base class representing a generic decorator for SiLA endpoints.

    Attributes:
      name: The name of the decorated object such as `Parameter` and `Response`.
      description: A description providing more details about the decorated object.
    """

    name: str = ""
    description: str = ""


def parse_structure(
    factory: type,
    feature: Feature,
    function: collections.abc.Callable,
    annotations: collections.abc.Sequence[type],
    decorators: collections.abc.Sequence["Decorator"],
    docs: collections.abc.Sequence,
) -> dict[str, Element]:
    """
    Parse the structure for parameters or (intermediate) responses by combining annotations, decorators, and docs.

    Args:
      factory: The factory class to generate (Parameters or Responses).
      feature: The SiLA feature context.
      function: The function for which the parameters or responses are parsed.
      annotations: The type annotations of the function parameters or return values.
      decorators: The decorators specifying additional metadata.
      docs: The parsed docstring metadata.

    Returns:
      The resulting structure (Parameters or (Intermediate-)Responses).

    Raises:
      TypeError: If a type cannot be parsed from the annotations.
    """

    if len(docs) > len(annotations):
        warnings.warn_explicit(
            f"More documented items than annotations in {function.__qualname__}, using only the first {len(annotations)}.",  # noqa: E501
            category=UserWarning,
            filename=inspect.getfile(inspect.unwrap(function)),
            lineno=inspect.getsourcelines(inspect.unwrap(function))[1],
        )

    if len(decorators) > len(annotations):
        warnings.warn_explicit(
            f"More decorators than annotations in {function.__qualname__}, using only the first {len(annotations)}.",
            category=UserWarning,
            filename=inspect.getfile(inspect.unwrap(function)),
            lineno=inspect.getsourcelines(inspect.unwrap(function))[1],
        )

    elements: dict[str, Element] = {}
    for index, (annotation, decorator, doc) in enumerate(
        itertools.islice(itertools.zip_longest(annotations, decorators, docs), len(annotations))
    ):
        name = None
        description = ""

        if decorator is not None:
            name = decorator.name or name
            description = decorator.description or description

        if doc is not None:
            name = doc.get("name", name)
            description = description or doc.get("default", description)  # prefer decorator description, if any

        try:
            data_type = infer(annotation, feature)
        except TypeError:
            import traceback

            traceback.print_exc()
            msg = f"Unable to identify SiLA type from annotation '{annotation.__name__}' in {function.__qualname__}."
            raise TypeError(msg) from None

        if name is None:
            name = f"Unnamed {index}"
            warnings.warn_explicit(
                f"No name found for {factory.__name__} {index} in {function.__qualname__}, defaulting to '{name}'.",
                category=UserWarning,
                filename=inspect.getfile(inspect.unwrap(function)),
                lineno=inspect.getsourcelines(inspect.unwrap(function))[1],
            )

        element = Element(
            identifier=name.replace(" ", ""),
            display_name=name,
            description=description,
            data_type=data_type,
        )
        elements[element.identifier] = element

    return elements



import dataclasses
import datetime

import typing_extensions as typing

from unitelabs.cdk import sila
from unitelabs.cdk.features.core import authorization_service

from ..authentication_service import AuthenticationService


class InvalidAccessToken(Exception):
    """The sent access token is not valid."""


@dataclasses.dataclass
class AccessToken(sila.Metadatum, errors=[InvalidAccessToken]):
    """Token to be sent with every call in order to get access to the SiLA Server functionality."""

    access_token: str

    @typing.override
    async def intercept(self) -> None:
        try:
            authentication_service = self._feature.app.get_feature(AuthenticationService)
            token = authentication_service.access_tokens[self.access_token]
        except KeyError:
            raise authorization_service.InvalidAccessToken from None
        else:
            token.last_usage = datetime.datetime.now()


class AuthorizationService(sila.Feature):
    """
    This Feature provides access control for the implementing server.

    It specifies the SiLA Client Metadata for the access token, that has been provided by the
    AuthenticationService core Feature.
    """

    def __init__(self, metadata: type[AccessToken] = AccessToken):
        super().__init__(originator="org.silastandard", category="core", metadata=[metadata])



import collections.abc
import dataclasses
import inspect

import typing_extensions as typing

from sila.server import Element, Feature

from .. import utils
from .decorator import Decorator, parse_structure

RESPONSES_ATTRIBUTE = "__responses"
DOCUMENTATION_KEYWORD = "return"


@dataclasses.dataclass
class Response(Decorator):
    """
    A decorator class for defining responses of an SiLA endpoint.

    Attributes:
      name: The name of the decorated response.
      description: A description providing more details about the decorated response.

    Examples:
      Option 1: Using decorators to define responses for an endpoint method:
      >>> @Response(name="Response A", description="A string response.")
      >>> @Response(name="Response B", description="An integer response.")
      >>> def example_handler(self) -> tuple[str, int]:
      >>>     ...

      Option 2: Using a docstring to define responses:
      >>> def example_handler(self) -> tuple[str, int]:
      >>>     \"\"\"
      >>>     .. return:: A string response.
      >>>       :name: Response A
      >>>     .. return:: An integer response.
      >>>       :name: Response B
      >>>     \"\"\"
      >>>     ...
    """

    def __call__(self, function: collections.abc.Callable) -> collections.abc.Callable:  # noqa: D102
        responses = getattr(function, RESPONSES_ATTRIBUTE, [])
        setattr(function, RESPONSES_ATTRIBUTE, [self, *responses])

        return function


@dataclasses.dataclass
class Responses:
    """Represents a SiLA structure containing response elements, inferred from a callable's signature."""

    elements: dict[str, Element] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_signature(
        cls,
        feature: Feature,
        function: collections.abc.Callable,
    ) -> typing.Self:
        """
        Infer and construct response elements based on function signature annotations, decorators, and documentation.

        Args:
          feature: The SiLA feature context.
          function: The function to analyze for response elements.

        Returns:
          A Responses object containing the inferred response elements.

        Raises:
          TypeError: If the function's return type annotation is invalid.
        """

        signature = inspect.signature(function)
        docs = utils.parse_docs(inspect.getdoc(function)).get(DOCUMENTATION_KEYWORD, [])
        decorators: list[Response] = getattr(function, RESPONSES_ATTRIBUTE, [])

        if signature.return_annotation is None or signature.return_annotation is inspect.Parameter.empty:
            return cls({})

        origin = typing.get_origin(signature.return_annotation) or signature.return_annotation
        return_annotation = (
            signature.return_annotation
            if origin is not typing.Annotated and issubclass(origin, tuple)
            else tuple[signature.return_annotation]
        )
        annotations = typing.get_args(return_annotation)

        return cls(
            parse_structure(
                Responses,
                feature,
                function,
                annotations,
                decorators,
                docs,
            )
        )



import typing
import unittest.mock

import pytest

import sila
from sila.framework.errors.defined_execution_error import DefinedExecutionError
from unitelabs.cdk.sila.command.observable_command import ObservableCommand
from unitelabs.cdk.sila.commands.responses import Response
from unitelabs.cdk.sila.common.feature import Feature


class DefinedError(Exception):
    pass


@pytest.fixture
def feature():
    return Feature(identifier="Feature", name="Feature")


@pytest.fixture
def handler():
    return ObservableCommand(identifier="ObservableCommand", name="Unobservable Command", errors=[DefinedError])


class TestAttach:
    async def test_should_attach_handler(self, feature: Feature, handler: ObservableCommand):
        # Attach
        attached = handler.attach(feature)

        # Assert that the method returns the correct value
        assert attached is True
        assert feature.commands[handler._identifier] == handler._handler

    async def test_should_ignore_disabled_handler(self, feature: Feature, handler: ObservableCommand):
        # Attach
        handler._enabled = False
        attached = handler.attach(feature)

        # Assert that the method returns the correct value
        assert attached is False
        assert handler._identifier not in feature.commands

    async def test_should_call_enabled_callback(self, feature: Feature, handler: ObservableCommand):
        # Attach
        handler._enabled = unittest.mock.Mock(return_value=True)
        attached = handler.attach(feature)

        # Assert that the method returns the correct value
        assert attached is True
        assert feature.commands[handler._identifier] == handler._handler
        handler._enabled.assert_called_once_with(feature)

    async def test_should_attach_return_annotation(self, feature: Feature, handler: ObservableCommand):
        # Create handler
        def function() -> int:
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        with pytest.warns(match="No name found for Responses 0"):
            handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.ObservableCommand)
        assert handler._handler.responses == {
            "Unnamed0": sila.Element(
                identifier="Unnamed0", display_name="Unnamed 0", description="", data_type=sila.Integer
            )
        }

    async def test_should_attach_missing_return_annotation(self, feature: Feature, handler: ObservableCommand):
        # Create handler
        def function():
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.ObservableCommand)
        assert handler._handler.responses == {}

    async def test_should_attach_annotated_return(self, feature: Feature, handler: ObservableCommand):
        # Create handler
        def function() -> typing.Annotated[int, sila.Unit("s", [sila.UnitComponent("Second")])]:
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        with pytest.warns(match="No name found for Responses 0"):
            handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.ObservableCommand)
        assert issubclass(handler._handler.responses["Unnamed0"].data_type, sila.Constrained)
        assert handler._handler.responses["Unnamed0"].data_type.data_type == sila.Integer
        assert handler._handler.responses["Unnamed0"].data_type.constraints == [
            sila.Unit("s", [sila.UnitComponent("Second")])
        ]


class TestExecute:
    # Execute synchronous function with default parameters.
    async def test_execute_synchronous_default_parameters(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        callback = unittest.mock.Mock()

        @Response("Response")
        def function() -> str:
            callback()
            return unittest.mock.sentinel.response

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(metadata={})

        # Assert that the function was called with the correct arguments
        callback.assert_called_once_with()

        # Assert that the method returns the correct value
        assert result == {"Response": unittest.mock.sentinel.response}

    # Execute synchronous function with custom parameters.
    async def test_execute_synchronous_custom_parameters(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        callback = unittest.mock.Mock()

        @Response("Response")
        def function(argument: str) -> str:
            callback(argument=argument)

            return unittest.mock.sentinel.response

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(metadata={}, argument="World, World!")

        # Assert that the function was called with the correct arguments
        callback.assert_called_once_with(argument="World, World!")

        # Assert that the method returns the correct value
        assert result == {"Response": unittest.mock.sentinel.response}

    # Execute synchronous function with no return value.
    async def test_execute_synchronous_returns_none(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        callback = unittest.mock.Mock()

        def function(argument: str):
            callback(argument=argument)

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(metadata={}, argument="World, World!")

        # Assert that the method returns the correct value
        callback.assert_called_once_with(argument="World, World!")
        assert result == {}

    # Verify that the method raises an error when the synchronous function raises.
    async def test_raises_when_synchronous_raises(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        def function():
            msg = "Hello, World!"
            raise Exception(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        with pytest.raises(Exception, match=r"Exception: Hello, World!"):
            await handler.execute(metadata={})

    # Verify that the method raises a defined execution error when the synchronous decorator knows the error type.
    async def test_raises_when_synchronous_raises_known_error(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        def function():
            msg = "Hello, World!"
            raise DefinedError(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        with pytest.raises(DefinedExecutionError) as exc_info:
            await handler.execute(metadata={})

        assert exc_info.value.identifier == "DefinedError"
        assert exc_info.value.display_name == "Defined Error"
        assert exc_info.value.description == "Common base class for all non-exit exceptions."
        assert exc_info.value.message == "Hello, World!"

    # Execute asynchronous function with default parameters.
    async def test_execute_asynchronous_default_parameters(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        callback = unittest.mock.AsyncMock()

        @Response("Response")
        async def function() -> str:
            await callback()
            return unittest.mock.sentinel.response

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(function=function, metadata={})

        # Assert that the function was called with the correct arguments
        callback.assert_awaited_once_with()

        # Assert that the method returns the correct value
        assert result == {"Response": unittest.mock.sentinel.response}

    # Execute asynchronous function with custom parameters.
    async def test_execute_asynchronous_custom_parameters(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        callback = unittest.mock.AsyncMock()

        @Response("Response")
        async def function(argument: str) -> str:
            await callback(argument=argument)

            return unittest.mock.sentinel.response

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(metadata={}, argument="World, World!")

        # Assert that the function was called with the correct arguments
        callback.assert_awaited_once_with(argument="World, World!")

        # Assert that the method returns the correct value
        assert result == {"Response": unittest.mock.sentinel.response}

    # Execute asynchronous function with no return value.
    async def test_execute_asynchronous_returns_none(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        callback = unittest.mock.AsyncMock()

        async def function(argument: str):
            await callback(argument=argument)

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(metadata={}, argument="World, World!")

        # Assert that the method returns the correct value
        callback.assert_awaited_once_with(argument="World, World!")
        assert result == {}

    # Verify that the method raises an error when the asynchronous function raises.
    async def test_raises_when_asynchronous_raises(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        async def function():
            msg = "Hello, World!"
            raise Exception(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        with pytest.raises(Exception, match=r"Exception: Hello, World!"):
            await handler.execute(metadata={})

    # Verify that the method raises a defined execution error when the asynchronous decorator knows the error type.
    async def test_raises_when_asynchronous_raises_known_error(self, feature: Feature, handler: ObservableCommand):
        # Initialize the observable command
        async def function():
            msg = "Hello, World!"
            raise DefinedError(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        with pytest.raises(DefinedExecutionError) as exc_info:
            await handler.execute(metadata={})

        assert exc_info.value.identifier == "DefinedError"
        assert exc_info.value.display_name == "Defined Error"
        assert exc_info.value.description == "Common base class for all non-exit exceptions."
        assert exc_info.value.message == "Hello, World!"



# ruff: noqa: D401

import time

from unitelabs.cdk import sila


class UnobservablePropertyTest(sila.Feature):
    """This feature tests a static and a dynamic unobservable property."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    @sila.UnobservableProperty()
    def get_answer_to_everything(self) -> int:
        """Returns the answer to the ultimate question of life, the universe, and everything. 42."""

        return 42

    @sila.UnobservableProperty()
    def get_seconds_since_1970(self) -> int:
        """Returns the unix timestamp: The time in seconds since January 1st of 1970."""

        return round(time.time())



from dataclasses import fields, is_dataclass

import typing_extensions as typing

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance


class InvalidSchemaFieldError(Exception):
    """The provided key is not present in the Schema."""


class Schema:
    """Wrapper for JSONSchema dictionaries."""

    def __init__(
        self,
        schema: dict[str, typing.Any],
        definitions: typing.Optional[dict[str, typing.Any]] = None,
    ):
        self._schema = schema
        self._schema_definitions = definitions or schema.get("$defs")

    @property
    def definition(self) -> dict[str, str]:
        """Get the JSONSchema definition."""
        return self._schema

    def get(self, field: str) -> "Schema":
        """
        Extract a nested `Schema` object for named `field`.

        Args:
          field: The name of the field in the schema to extract.

        Raises:
          InvalidSchemaKeyError: If the provided `field` is not present in the schema.
        """
        if not (schema := self._schema["properties"].get(field, {})):
            valid_fields = list(self._schema["properties"].keys())
            msg = f"{field} is not a valid field or nested schema, valid schemas include: {valid_fields}"
            raise InvalidSchemaFieldError(msg)

        if ref := schema.get("$ref", "").removeprefix("#/$defs/"):
            return Schema(self._schema_definitions[ref], self._schema_definitions)

        return Schema(schema, self._schema_definitions)


def describe(
    dataclass: type["DataclassInstance"], schema: Schema, default: typing.Optional["DataclassInstance"] = None
) -> dict[str, typing.Any]:
    """
    Build a dict of values describing the keys, their types, descriptions and defaults.

    Args:
      dataclass: The dataclass to describe.
      schema: A wrapped JSONSchema for the dataclass, used for extracting description docstrings.
      default: A default instance of the dataclass to use for extracting default values.
    """
    description = {}
    default = default or dataclass()
    for field in fields(dataclass):
        name = field.name
        field_type = field.type
        field_schema = schema.get(name).definition
        field_default = getattr(default, name)
        data = dict(
            type=get_type_str(field_type),
            description=field_schema.get("description", "").replace("\n", " "),
            default=f"'{field_default}'" if isinstance(field_default, str) else str(field_default),
        )

        if is_dataclass(field_type):
            data.update(
                default="",
                description=f"For more information use `config show --field {name}`.",
                values=describe(field_type, schema.get(name), field_default),
            )

        description[name] = data

    return description


def get_type_str(type_: typing.Any) -> str:  # noqa: ANN401
    """
    Get a formatted string representation of a type.

    Args:
      type_: The type to get a string representation of.

    Returns:
      A formatted string representation of the type.

    Examples:
      >>> get_type_str(int)
      'int'
      >>> get_type_str(list[int])
      'list[int]'
      >>> get_type_str(list[typing.Union[int, str]])
      'list[int | str]'
      >>> get_type_str(list[dataclass_type])
      'list[dataclass_type]'
    """
    origin = typing.get_origin(type_)
    if args := typing.get_args(type_):
        import typing as t

        if origin == typing.Annotated:
            args = args[:-1]
        if origin in [t.Literal, typing.Literal]:
            args = (
                [f"'{arg}'" for arg in args]
                if all(isinstance(arg, str) for arg in args)
                else [str(arg) for arg in args]
            )
        if origin in [list, set, dict]:
            return f"{origin.__name__}[{', '.join([get_type_str(arg) for arg in args])}]"
        return " | ".join([get_type_str(arg) for arg in args])

    return getattr(type_, "__name__", type_)



import abc

from unitelabs.cdk import sila


class StartSimulationModeFailed(Exception):
    """
    The server cannot change to Simulation Mode.

    This error can, e.g., be thrown, if a real-world process needs to be ended before switching to simulation mode.
    """


class StartRealModeFailed(Exception):
    """
    The server cannot change to Real Mode.

    This error can, e.g., be thrown, if a device is not ready to change into Real Mode.
    """


class SimulationControllerBase(sila.Feature, metaclass=abc.ABCMeta):
    """
    This Feature provides control over the simulation behaviour of a SiLA Server.

    A SiLA Server can run in two modes:
    (a) Real Mode - with real activities, e.g. addressing or controlling real hardware,
        e.g. through serial/CANBus commands, writing to real databases, moving real objects etc.
    (b) Simulation Mode - where every command is only simulated and responses are just example returns.

    Note that certain commands and properties might not be affected by this feature if they
    do not interact with the real world.
    """

    _simulation_mode = False

    def __init__(self):
        super().__init__(
            originator="org.silastandard",
            category="core",
            version="1.0",
            maturity_level="Verified",
        )

    @abc.abstractmethod
    @sila.UnobservableCommand(errors=[StartSimulationModeFailed])
    async def start_simulation_mode(self) -> None:
        """
        Set the SiLA Server to run in Simulation Mode, i.e. all following commands are executed in simulation mode.

        The Simulation Mode can only be entered, if all hardware operations have been safely terminated
        or are in a controlled, safe state.

        The simulation mode can be stopped by issuing the 'Start Real Mode' command.
        """

    @abc.abstractmethod
    @sila.UnobservableCommand(errors=[StartRealModeFailed])
    async def start_real_mode(self) -> None:
        """
        Set the SiLA Server to run in real mode.

        In real-mode all following commands are executed with real-world
        interactions, like serial port/CAN communication, motor actions etc.

        If the server is in Simulation Mode it can be interrupted at any time. A re-initialization of
        the hardware might be required. The Real Mode can be stopped by issuing the 'Start Simulation Mode' command.
        """

    @abc.abstractmethod
    @sila.UnobservableProperty(name="SimulationMode")
    async def simulation_mode(self) -> bool:
        """Whether or not the SiLA Server is in Simulation Mode."""



# ruff: noqa: D205, D401, E501

import dataclasses
import datetime
import uuid

import typing_extensions as typing

from sila.server import FeatureIdentifier
from unitelabs.cdk import sila


class AuthenticationFailed(Exception):
    """The provided credentials are not valid."""


class InvalidAccessToken(Exception):
    """The sent access token is not valid."""


@dataclasses.dataclass
class AccessToken:
    """An access token used for authorization."""

    token: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    """The token used by the client for authorization."""

    scope: list[FeatureIdentifier] = dataclasses.field(default_factory=list)
    """The requested scope of access."""

    lifetime: datetime.timedelta = dataclasses.field(default_factory=datetime.timedelta)
    """The lifetime of the access token before it expires."""

    last_usage: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.now)
    """Point in time when the access token was used last."""

    @property
    def is_expired(self) -> bool:
        """Whether the access token has expired."""

        return (datetime.datetime.now() - self.lifetime) > self.last_usage


class AuthenticationService(sila.Feature):
    """
    This Feature provides SiLA Clients with access tokens based on a user identification and password.
    1. the user needs to login with the Login command into the server with a user identification (=user name) and a password
    2. after verification, an Access Token with the Token Lifetime information will be generated and provided by the server.
    3. the user can log-out from the server with the Logout command - a valid Access Token is required to run this command.
    """

    def __init__(self):
        super().__init__(originator="org.silastandard", category="core")

        self.access_tokens = dict[str, AccessToken]()
        self.default_lifetime = datetime.timedelta(hours=1)

    @sila.UnobservableCommand(errors=[AuthenticationFailed])
    @sila.Response("Access Token")
    @sila.Response("Token Lifetime")
    async def login(
        self,
        user_identification: str,
        password: str,
        requested_server: typing.Annotated[
            str,
            sila.constraints.Length(36),
            sila.constraints.Pattern(r"[0-9a-f]{8}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{12}"),
        ],
        requested_features: list[
            typing.Annotated[
                str,
                sila.constraints.FullyQualifiedIdentifier("FeatureIdentifier"),
            ]
        ],
    ) -> tuple[
        str,
        typing.Annotated[int, sila.constraints.Unit("s", components=[sila.constraints.UnitComponent("Second")])],
    ]:
        """
        Provides an access token based on user information.

        .. parameter:: The user identification string (e.g. a user name)
        .. parameter:: The password
        .. parameter:: The ServerUUID of the server for which an authorization is requested.
        .. parameter:: The fully qualified identifiers of features that are requested to access. If no feature is provided, this means that all features are requested.
        .. return:: The token to be used along with accessing a Command or Property on a SiLA Server.
        .. return:: The lifetime (in seconds) of the provided access token as the maximum validity period after the last SiLA Server request.
        """

        if requested_server != self.server.uuid:
            msg = f"Requested access to server '{requested_server}' which could not be granted from server '{self.server.uuid}'."
            raise AuthenticationFailed(msg)

        features = list[FeatureIdentifier]()
        for requested_feature in requested_features:
            feature_identifier = FeatureIdentifier(requested_feature)
            features.append(feature_identifier)
            if feature_identifier not in self.server.features:
                msg = f"Requested access to unknown feature '{feature_identifier}' which could not be granted."
                raise AuthenticationFailed(msg)

        if not self.validate(user_identification, password, features):
            msg = "Received invalid credentials."
            raise AuthenticationFailed(msg)

        token = AccessToken(scope=features, lifetime=self.default_lifetime)
        self.access_tokens[token.token] = token

        return (token.token, int(token.lifetime.total_seconds()))

    @sila.UnobservableCommand(errors=[InvalidAccessToken])
    async def logout(self, access_token: str) -> None:
        """
        Invalidates the given access token immediately.

        .. parameter:: The access token to be invalidated
        """

        try:
            self.access_tokens.pop(access_token)
        except KeyError:
            msg = "Received invalid or unknown token."
            raise InvalidAccessToken(msg) from None

    async def validate(self, username: str, password: str, scope: list[FeatureIdentifier]) -> bool:
        """
        Subclass to validate the given credentials.

        Args:
          username: The username.
          password: The password.
          scope: The list requested features.

        Returns:
          Whether the given username and password valid and authorized to
            access the requested scope.
        """

        raise NotImplementedError



import dataclasses

import typing_extensions as typing

from sila.server import Custom

from ..common import Dataclass, Feature


@dataclasses.dataclass
class CustomDataType(Dataclass):
    """A SiLA custom data type definition."""

    @typing.override
    @classmethod
    def attach(cls, feature: Feature) -> type[Custom]:
        super().attach(feature)

        if cls._identifier in feature.data_type_definitions:
            return feature.data_type_definitions[cls._identifier]

        cls._custom = Custom.create(
            identifier=cls._identifier,
            display_name=cls._name,
            description=cls._description,
            data_type=cls._infer_data_type(feature),
            feature=feature,
        )
        feature._custom_data_types[cls._identifier] = cls
        return cls._custom



# ruff: noqa: D205, D401

import typing

from sila.server import Serializer
from unitelabs.cdk import sila


class UnimplementedFeature(Exception):
    """The Feature specified by the given Feature identifier is not implemented by the server."""


class SiLAService(sila.Feature):
    """
    This Feature MUST be implemented by each SiLA Server.

    It specifies Commands and Properties to discover the Features a SiLA Server implements as well as details
    about the SiLA Server, like name, type, description, vendor and UUID.

    Any interaction described in this feature MUST not affect the behaviour of any other Feature.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            originator="org.silastandard",
            category="core",
            version="1.0",
            maturity_level="Normative",
            name="SiLA Service",
        )

    @sila.UnobservableCommand(errors=[UnimplementedFeature])
    @sila.Response(name="Feature Definition")
    def get_feature_definition(
        self,
        feature_identifier: typing.Annotated[str, sila.constraints.FullyQualifiedIdentifier("FeatureIdentifier")],
    ) -> typing.Annotated[
        str,
        sila.constraints.Schema(
            "Xml",
            url="https://gitlab.com/SiLA2/sila_base/raw/master/schema/FeatureDefinition.xsd",
        ),
    ]:
        """
        Get the Feature Definition of an implemented Feature by its fully qualified Feature Identifier.
        This command has no preconditions and no further dependencies and can be called at any time.

        .. parameter:: The fully qualified Feature identifier for which the Feature definition shall be retrieved.
        .. return:: The Feature definition in XML format (according to the Feature Definition Schema).
        """

        try:
            feature = self.server.get_feature(feature_identifier)
        except ValueError:
            raise UnimplementedFeature from None

        serializer = Serializer()
        feature.serialize(serializer)

        return serializer.buffer.getvalue()

    @sila.UnobservableCommand()
    def set_server_name(self, server_name: typing.Annotated[str, sila.constraints.MaximalLength(255)]) -> None:
        """
        Sets a human readable name to the Server Name Property.Command has no preconditions and
        no further dependencies and can be called at any time.

        .. parameter:: The human readable name to assign to the SiLA Server.
        """

        self.server.name = server_name

    @sila.UnobservableProperty(name="Server UUID")
    def get_server_uuid(
        self,
    ) -> typing.Annotated[
        str,
        sila.constraints.Length(36),
        sila.constraints.Pattern(r"[0-9a-f]{8}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{12}"),
    ]:
        """
        Globally unique identifier that identifies a SiLA Server. The Server UUID MUST be generated once
        and remain the same for all times.
        """

        return self.server.uuid

    @sila.UnobservableProperty()
    def get_server_name(self) -> typing.Annotated[str, sila.constraints.MaximalLength(255)]:
        """Human readable name of the SiLA Server. The name can be set using the 'Set Server Name' command."""

        return self.server.name

    @sila.UnobservableProperty()
    def get_server_type(self) -> typing.Annotated[str, sila.constraints.Pattern(r"[A-Z][a-zA-Z0-9]*")]:
        """
        The type of this server. It, could be, e.g., in the case of a SiLA Device the model name.
        It is specified by the implementer of the SiLA Server and MAY not be unique.
        """

        return self.server.type

    @sila.UnobservableProperty()
    def get_server_description(self) -> str:
        """
        Get the description of the SiLA Server.

        This should include the use and purpose of this SiLA Server.
        """

        return self.server.description

    @sila.UnobservableProperty()
    def get_server_version(
        self,
    ) -> typing.Annotated[
        str, sila.constraints.Pattern(r"(0|[1-9][0-9]*)\.(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))?(_[_a-zA-Z0-9]+)?")
    ]:
        """
        Returns the version of the SiLA Server. A "Major" and a "Minor" version number (e.g. 1.0) MUST be provided,
        a Patch version number MAY be provided. Optionally, an arbitrary text, separated by an underscore MAY be
        appended, e.g. “3.19.373_mighty_lab_devices”.
        """

        return self.server.version

    @sila.UnobservableProperty(name="Server Vendor URL")
    def get_server_vendor_url(self) -> typing.Annotated[str, sila.constraints.Pattern(r"https?://.+")]:
        """
        Returns the URL to the website of the vendor or the website of the product of this SiLA Server.
        This URL SHOULD be accessible at all times.
        The URL is a Uniform Resource Locator as defined in RFC 1738.
        """

        return self.server.vendor_url

    @sila.UnobservableProperty()
    def get_implemented_features(
        self,
    ) -> list[typing.Annotated[str, sila.constraints.FullyQualifiedIdentifier("FeatureIdentifier")]]:
        """
        Returns a list of fully qualified Feature identifiers of all implemented Features of this SiLA Server.
        This list SHOULD remain the same throughout the lifetime of the SiLA Server.
        """

        return [str(feature_identifier) for feature_identifier in self.server.features]



from .decorator import Decorator
from .intermediate_responses import IntermediateResponse, IntermediateResponses
from .parameters import Parameter, Parameters
from .responses import Response, Responses

__all__ = [
    "Decorator",
    "IntermediateResponse",
    "IntermediateResponses",
    "Parameter",
    "Parameters",
    "Response",
    "Responses",
]



import collections.abc
import dataclasses
import inspect

import typing_extensions as typing

from sila.server import Element, Feature

from .. import utils
from .decorator import Decorator, parse_structure
from .intermediate import Intermediate

INTERMEDIATE_RESPONSES_ATTRIBUTE = "__intermediate_responses"


@dataclasses.dataclass
class IntermediateResponse(Decorator):
    '''
    A decorator class for defining intermediate responses of an SiLA endpoint.

    Attributes:
      name: The name of the decorated response.
      description: A description providing more details about the decorated response.

    Examples:
      Option 1: Using decorators to define intermediate responses for an endpoint method:
      >>> @Response(name="Intermediate Response A", description="A string intermediate response.")
      >>> @Response(name="Intermediate Response B", description="An integer intermediate response.")
      >>> def example_handler(self, /, *, intermediate: Intermediate[str, int]):
      >>>     ...

      Option 2: Using a docstring to define intermediate responses:
      >>> def example_handler(self, /, *, intermediate: Intermediate[str, int]):
      >>>     """
      >>>     .. yield:: A string intermediate response.
      >>>       :name: Intermediate Response A
      >>>     .. yield:: An integer intermediate response.
      >>>       :name: Intermediate Response B
      >>>     """
      >>>     ...
    '''

    def __call__(self, function: collections.abc.Callable) -> collections.abc.Callable:  # noqa: D102
        intermediate_responses = getattr(function, INTERMEDIATE_RESPONSES_ATTRIBUTE, [])
        setattr(function, INTERMEDIATE_RESPONSES_ATTRIBUTE, [self, *intermediate_responses])

        return function


@dataclasses.dataclass
class IntermediateResponses:
    """Represents a SiLA structure containing intermediate response elements, inferred from a callable's signature."""

    elements: dict[str, Element] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_signature(
        cls,
        feature: Feature,
        function: collections.abc.Callable,
    ) -> typing.Self:
        """
        Infer and construct intermediate response elements.

        Construction is based on function signature annotations, decorators, and documentation.

        Args:
          feature: The SiLA feature context.
          function: The function to analyze for intermediate response elements.

        Returns:
          An IntermediateResponses object containing the inferred intermediate response elements.

        Raises:
          TypeError: If the function's `intermediate` parameter type annotation is missing or invalid.
        """

        signature = inspect.signature(function)
        docs = utils.parse_docs(inspect.getdoc(function)).get("yield", [])
        decorators: list[IntermediateResponse] = getattr(function, INTERMEDIATE_RESPONSES_ATTRIBUTE, [])

        intermediate_parameter = signature.parameters.get("intermediate", None)

        if intermediate_parameter is None:
            return cls({})

        if typing.get_origin(intermediate_parameter.annotation) is not Intermediate:
            msg = f"Missing type annotation for parameter 'intermediate' in {function.__qualname__}."
            raise TypeError(msg)

        intermediate_annotation = typing.get_args(intermediate_parameter.annotation)[0]
        intermediate_annotation = (
            intermediate_annotation
            if issubclass(typing.get_origin(intermediate_annotation) or intermediate_annotation, tuple)
            else tuple[intermediate_annotation]
        )
        annotations = typing.get_args(intermediate_annotation)

        return cls(
            parse_structure(
                IntermediateResponses,
                feature,
                function,
                annotations,
                decorators,
                docs,
            )
        )



import collections.abc
import functools
import inspect
import warnings
import weakref

import typing_extensions as typing

from sila import Element, MetadataIdentifier, Structure
from sila.server import CommandExecution, Handler, Native, SiLAError, UndefinedExecutionError

from ..data_types import from_sila, to_sila
from ..metadata import Metadatum
from .errors import define_error

if typing.TYPE_CHECKING:
    from .feature import Feature


class Decorator:
    """Base class for decorator based SiLA annotations."""

    def __init__(
        self,
        /,
        *,
        identifier: typing.Optional[str] = None,
        display_name: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        errors: typing.Optional[collections.abc.Sequence[type[Exception]]] = None,
        enabled: typing.Union[bool, collections.abc.Callable[..., bool]] = True,
    ) -> None:
        if display_name is not None:
            msg = "Using `display_name` is deprecated, use `name` instead."
            warnings.warn(msg, stacklevel=2)
            name = display_name

        self._identifier = identifier or ""
        self._name = name or ""
        self._description = ""
        self._enabled = enabled
        self._parameters: dict[str, Element] = {}
        self._responses: dict[str, Element] = {}
        self._intermediate_responses: dict[str, Element] = {}
        self._handler: typing.Optional[Handler] = None
        self._metadata: tuple[str, list[type[Metadatum]]] = ("", [])
        self._errors: list[type[Exception]] = list(errors or [])
        self._function: collections.abc.Callable = lambda: ...
        self._feature: typing.Optional["Feature"] = None

    def __call__(self, function: collections.abc.Callable) -> collections.abc.Callable:
        """Call the decorator to wrap the given method."""

        self._function = weakref.proxy(function)
        setattr(function, "__handler", self)

        return function

    def is_enabled(self, feature: "Feature") -> bool:
        """Whether the handler is enabled or not."""

        if callable(self._enabled):
            return self._enabled(feature)

        return self._enabled

    def attach(self, feature: "Feature") -> bool:
        """
        Create and attach a handler to the `feature`.

        Args:
          feature: The `Feature` to which the handler will be attached.

        Returns:
          Whether the handler was attached or not.
        """

        if not self.is_enabled(feature):
            return False

        self._feature = weakref.proxy(feature)
        return True

    async def execute(self, metadata: dict[MetadataIdentifier, Native], **parameters) -> Native:
        """
        Execute a given function with the provided keyword arguments.

        Args:
          metadata: Additional metadata sent from client to server.

        Returns:
          The result of the `function` execution.

        Raises:
          DefinedExecutionError: If the error type is in the list of defined errors.
          UndefinedExecutionError: If an unexpected error occurs during execution.
        """

        if not self._feature:
            raise RuntimeError

        try:
            function = self._with_metadata(self._function, metadata)
            function = self._with_parameters(function, parameters)

            responses = await self._execute(function)

            return to_sila(responses, Structure.create(self._responses), self._feature)
        except SiLAError:
            raise
        except Exception as error:
            import traceback

            traceback.print_exc()
            if type(error) in self._errors:
                raise define_error(error)(str(error)) from None

            msg = f"{error.__class__.__name__}: {error}"
            raise UndefinedExecutionError(msg) from error

    async def _execute(self, function: collections.abc.Callable) -> Native:
        responses = function()

        if inspect.isawaitable(responses):
            responses = await responses

        if responses is None:
            return {}

        return responses

    def _with_metadata(
        self, function: collections.abc.Callable, metadata: dict[MetadataIdentifier, Native]
    ) -> collections.abc.Callable:
        if self._metadata[0]:
            function = functools.partial(
                function,
                **{
                    self._metadata[0]: {
                        metadatum: metadatum.from_native(metadata[metadatum._metadatum.fully_qualified_identifier()])
                        for metadatum in self._metadata[1]
                        if metadatum._metadatum
                    }
                },
            )

        return function

    def _with_parameters(
        self,
        function: collections.abc.Callable,
        parameters: collections.abc.Mapping[str, typing.Union[Native, CommandExecution]],
    ) -> collections.abc.Callable:
        if not self._feature:
            raise RuntimeError

        parameters = dict(parameters)
        command_execution = parameters.pop("command_execution", None)
        parameters = typing.cast(dict[str, Native], parameters)

        if command_execution and isinstance(command_execution, CommandExecution):
            from ..commands.intermediate import Intermediate
            from ..commands.status import Status

            signature = inspect.signature(function)
            for param in signature.parameters.values():
                annotation = typing.get_origin(param.annotation) or param.annotation
                if annotation is Status:
                    function = functools.partial(function, **{param.name: Status(command_execution)})
                if annotation is Intermediate:
                    function = functools.partial(
                        function, **{param.name: Intermediate(command_execution, self._intermediate_responses)}
                    )

        return functools.partial(
            function, **from_sila(parameters, Structure.create(elements=self._parameters), self._feature)
        )



import dataclasses
import json

import pydantic
import typing_extensions as typing


@dataclasses.dataclass
class Nested:
    value: int = -1


@dataclasses.dataclass
class Base:
    __pydantic_config__ = pydantic.ConfigDict(
        validate_assignment=True,
        revalidate_instances="always",
        use_attribute_docstrings=True,
    )
    value: str = ""
    nested: Nested = dataclasses.field(default_factory=Nested)

    @classmethod
    def to_pydantic_dataclass(cls) -> type[typing.Self]:
        """Create a pydantic dataclass from the Config."""
        subclass = type(cls.__name__, (cls,), {})
        return pydantic.dataclasses.dataclass(subclass)

    @staticmethod
    def get_default() -> dict:
        """Get the default dictionary representation of the config."""
        configs = [b for b in Base.__subclasses__() if b.__name__ != "Base"]
        if not configs:
            raise ValueError
        config = configs[0]
        type_adapter = pydantic.TypeAdapter(config)
        return json.loads(type_adapter.dump_json(config()))

    @classmethod
    def load(cls, data: dict) -> typing.Self:
        """This docstring breaks everything ?

        Args:
          data: the data to load into the cls.
        """
        return cls.to_pydantic_dataclass()(**data)


# data = {"value": "a", "nested": {"value": 0}}
try:

    @dataclasses.dataclass
    class Derived(Base):
        more: str = "yes"

        @pydantic.field_validator("simple")
        @classmethod
        def must_be_true(cls, value: bool) -> bool:
            if not value:
                msg = "simple must be True."
                raise ValueError(msg)
            return value

    data = Derived.get_default()
    loaded = Derived.load(data)
    print(loaded)
except SyntaxError:
    print("how?")
else:
    print("alles gut.")



# ruff: noqa: D401, E501

import typing_extensions as typing

from unitelabs.cdk import sila

from ..metadata_provider import StringMetadata
from ..metadata_provider.metadata_provider import TwoIntegersMetadata


class MetadataConsumerTest(sila.Feature):
    """Provides commands and properties to set or respectively get SiLA Any Type values via command parameters or property responses respectively."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    @sila.UnobservableCommand()
    @sila.Response(name="Received String Metadata")
    async def echo_string_metadata(self, *, metadata: typing.Annotated[sila.Metadata, StringMetadata]) -> str:
        """
        Expects the "String Metadata" metadata from the "Metadata Provider" feature and responds with the metadata value.

        .. return:: The received string metadata
        """

        return metadata[StringMetadata].string_metadata

    @sila.UnobservableCommand()
    @sila.Response(name="Received String")
    @sila.Response(name="First Received Integer")
    @sila.Response(name="Second Received Integer")
    async def unpack_metadata(
        self, *, metadata: typing.Annotated[sila.Metadata, StringMetadata, TwoIntegersMetadata]
    ) -> tuple[str, int, int]:
        """
        Expects the "String Metadata" and "Two Integers Metadata" metadata from the "Metadata Provider" feature and responds with all three data items.

        .. return:: The received string (via "String Metadata")
        .. return:: The first element of the received integer structure (via "Two Integers Metadata")
        .. return:: The second element of the received integer structure (via "Two Integers Metadata")
        """

        return (
            metadata[StringMetadata].string_metadata,
            metadata[TwoIntegersMetadata].first_integer,
            metadata[TwoIntegersMetadata].second_integer,
        )

    @sila.UnobservableProperty()
    async def received_string_metadata(self, *, metadata: typing.Annotated[sila.Metadata, StringMetadata]) -> str:
        """Expects the "String Metadata" metadata from the "Metadata Provider" feature and returns the metadata value."""

        return metadata[StringMetadata].string_metadata

    @sila.ObservableProperty()
    async def received_string_metadata_as_characters(
        self, *, metadata: typing.Annotated[sila.Metadata, StringMetadata]
    ) -> sila.Stream[typing.Annotated[str, sila.constraints.Length(1)]]:
        """Expects the "String Metadata" metadata from the "Metadata Provider" feature and returns all characters of its string value as separate responses."""

        for char in metadata[StringMetadata].string_metadata:
            yield char



import itertools

import pytest

from unitelabs.cdk.sila.utils import parse_version


versions = [
    # Final releases
    pytest.param("0.0.0", "0.0", id="final"),
    pytest.param("1.0.0", "1.0", id="final"),
    pytest.param("1.2.3", "1.2.3", id="final"),
    pytest.param("1.2.03", "1.2.3", id="final"),
    pytest.param("v1.2.3", "1.2.3", id="final"),
    pytest.param("2012.10", "2012.10", id="date-based"),
    # Pre-releases
    *[
        pytest.param("".join(v), "1.2.3_a4", id="alpha")
        for v in itertools.product(("1.2.3",), ("", ".", "-", "_"), ("a", "alpha"), ("", ".", "-", "_"), ("4",))
    ],
    *[
        pytest.param("".join(v), "1.2.3_b4", id="beta")
        for v in itertools.product(("1.2.3",), ("", ".", "-", "_"), ("b", "beta"), ("", ".", "-", "_"), ("4",))
    ],
    *[
        pytest.param("".join(v), "1.2.3_rc4", id="rc")
        for v in itertools.product(
            ("1.2.3",), ("", ".", "-", "_"), ("c", "rc", "pre", "preview"), ("", ".", "-", "_"), ("4",)
        )
    ],
    # Post-releases
    *[
        pytest.param("".join(v), "1.2.3_post4", id="post")
        for v in itertools.product(("1.2.3",), ("", ".", "-", "_"), ("post", "rev", "r"), ("", ".", "-", "_"), ("4",))
    ],
    # Developmental releases
    *[
        pytest.param("".join(v), "1.2.3_dev4", id="dev")
        for v in itertools.product(("1.2.3",), ("", ".", "-", "_"), ("dev",), ("", ".", "-", "_"), ("4",))
    ],
]


class TestParseVersion:
    @pytest.mark.parametrize("python_version,sila_version", versions)
    def test_should_parse_version(self, python_version, sila_version):
        assert parse_version(python_version) == sila_version

    def test_should_raise_on_invalid_version(self):
        with pytest.raises(ValueError, match=r"Invalid version format: 'test'."):
            parse_version("test")



import collections.abc
import dataclasses
import datetime

import typing_extensions as typing

import sila
import sila.datetime
from sila import (
    Binary,
    Boolean,
    Constrained,
    Custom,
    DataType,
    Date,
    Integer,
    List,
    Native,
    Real,
    String,
    Structure,
    Time,
    Timestamp,
)

if typing.TYPE_CHECKING:
    from ..common.feature import Feature
    from .custom_data_type import CustomDataType


Any = typing.Union[
    str,
    int,
    float,
    bytes,
    bool,
    sila.datetime.date,
    sila.datetime.time,
    sila.datetime.datetime,
    collections.abc.Sequence["Any"],
    collections.abc.Mapping[str, "Any"],
    "CustomDataType",
]


def to_sila(
    value: Native,
    data_type: type[DataType],
    feature: "Feature",
) -> sila.Native:
    """
    Convert a CDK native value to a SiLA native value.

    Args:
      value: The native CDK value to convert.
      data_type: The SiLA data type of the message.
      feature: The feature containing custmo data types.

    Returns:
      The converted SiLA native value.
    """

    if issubclass(data_type, Custom):
        custom = feature._custom_data_types[data_type.identifier]
        if not isinstance(value, custom):
            msg = f"Expected value to be of type '{custom.__name__}', received {type(value)}."
            raise ValueError(msg)

        fields = dataclasses.fields(custom)
        values = {field.name: getattr(value, field.name) for field in fields}

        if custom._custom and issubclass(custom._custom.data_type, Structure):
            return to_sila(values, data_type.data_type, feature)

        field = fields[0]
        item = values.get(field.name, None)

        if item is None:
            msg = f"Missing item '{field.name}' in values {values}."
            raise ValueError(msg)

        return to_sila(item, data_type.data_type, feature)

    if issubclass(data_type, Structure):
        if not isinstance(value, tuple) and not isinstance(value, dict):
            value = (value,)

        if isinstance(value, tuple):
            values = {}
            for index, name in enumerate(data_type.elements.keys()):
                if len(value) < index:
                    msg = f"Expected {len(data_type.elements)} elements in tuple, received {value}."
                    raise ValueError(msg)
                values[name] = value[index]

            value = values

        if not isinstance(value, dict):
            msg = f"Expected value to be of type 'dict', received {type(value)}."
            raise ValueError(msg)

        return {name: to_sila(value[name], element.data_type, feature) for name, element in data_type.elements.items()}

    if issubclass(data_type, List):
        if not isinstance(value, list):
            msg = f"Expected value to be of type 'list', received {type(value)}."
            raise ValueError(msg)

        return [to_sila(item, data_type.data_type, feature) for item in value]

    if issubclass(data_type, Constrained):
        return to_sila(value, data_type.data_type, feature)

    return value


@typing.overload
def from_sila(
    value: dict[str, sila.Native], data_type: type[Structure], feature: "Feature"
) -> collections.abc.Mapping[str, Any]: ...


@typing.overload
def from_sila(value: list[sila.Native], data_type: type[List], feature: "Feature") -> collections.abc.Sequence[Any]: ...


@typing.overload
def from_sila(value: sila.Native, data_type: type[Constrained], feature: "Feature") -> Any: ...


@typing.overload
def from_sila(value: sila.Native, data_type: type[Custom], feature: "Feature") -> "CustomDataType": ...


@typing.overload
def from_sila(value: Native, data_type: type[sila.Any], feature: "Feature") -> Any: ...


@typing.overload
def from_sila(value: str, data_type: type[String], feature: "Feature") -> str: ...


@typing.overload
def from_sila(value: int, data_type: type[Integer], feature: "Feature") -> int: ...


@typing.overload
def from_sila(value: float, data_type: type[Real], feature: "Feature") -> float: ...


@typing.overload
def from_sila(value: bytes, data_type: type[Binary], feature: "Feature") -> bytes: ...


@typing.overload
def from_sila(value: bool, data_type: type[Boolean], feature: "Feature") -> bool: ...


@typing.overload
def from_sila(value: datetime.date, data_type: type[Date], feature: "Feature") -> sila.datetime.date: ...


@typing.overload
def from_sila(value: datetime.time, data_type: type[Time], feature: "Feature") -> sila.datetime.time: ...


@typing.overload
def from_sila(value: datetime.datetime, data_type: type[Timestamp], feature: "Feature") -> sila.datetime.datetime: ...


@typing.overload
def from_sila(
    value: sila.Native, data_type: type[DataType], feature: "Feature"
) -> typing.Union[dict[str, Any], list[Any], Any]: ...


def from_sila(
    value: sila.Native, data_type: type[DataType], feature: "Feature"
) -> typing.Union[collections.abc.Mapping[str, Any], collections.abc.Sequence[Any], Any]:
    """
    Convert a SiLA native value to a CDK native value.

    Args:
      value: The native SiLA value to convert.
      data_type: The SiLA data type of the message.
      feature: The feature containing custmo data types.

    Returns:
      The converted CDK native value.
    """

    if issubclass(data_type, Custom):
        custom = feature._custom_data_types[data_type.identifier]

        if custom._custom and issubclass(custom._custom.data_type, Structure):
            if not isinstance(value, dict):
                msg = f"Expected value to be of type 'dict', received {type(value)}."
                raise ValueError(msg)

            return custom(**from_sila(value, custom._custom.data_type, feature))

        field = dataclasses.fields(custom)[0]
        return custom(**{field.name: from_sila(value, custom._custom.data_type, feature)})

    if issubclass(data_type, Structure):
        if not isinstance(value, dict):
            msg = f"Expected value to be of type 'dict', received {type(value)}."
            raise ValueError(msg)

        return {
            name: from_sila(value[name], element.data_type, feature) for name, element in data_type.elements.items()
        }

    if issubclass(data_type, List):
        if not isinstance(value, list):
            msg = f"Expected value to be of type 'list', received {type(value)}."
            raise ValueError(msg)

        return [from_sila(item, data_type.data_type, feature) for item in value]

    if issubclass(data_type, Constrained):
        return from_sila(value, data_type.data_type, feature)

    if issubclass(data_type, Date):
        if isinstance(value, sila.datetime.date):
            return value

        if isinstance(value, datetime.date):
            return sila.datetime.date(year=value.year, month=value.month, day=value.day)

        msg = f"Expected value to be of type 'date', received {type(value)}."
        raise ValueError(msg)

    if issubclass(data_type, Time):
        if isinstance(value, sila.datetime.time):
            return value

        if isinstance(value, datetime.time):
            return sila.datetime.time(
                hour=value.hour,
                minute=value.minute,
                second=value.second,
                microsecond=value.microsecond,
                tzinfo=value.tzinfo,
            )

        msg = f"Expected value to be of type 'time', received {type(value)}."
        raise ValueError(msg)

    if issubclass(data_type, Timestamp):
        if isinstance(value, sila.datetime.datetime):
            return value

        if isinstance(value, datetime.datetime):
            return sila.datetime.datetime(
                year=value.year,
                month=value.month,
                day=value.day,
                hour=value.hour,
                minute=value.minute,
                second=value.second,
                microsecond=value.microsecond,
                tzinfo=value.tzinfo,
            )

        msg = f"Expected value to be of type 'time', received {type(value)}."
        raise ValueError(msg)

    if issubclass(data_type, String):
        if not isinstance(value, str):
            msg = f"Expected value to be of type 'str', received {type(value)}."
            raise ValueError(msg)

        return value

    if issubclass(data_type, Integer):
        if not isinstance(value, int):
            msg = f"Expected value to be of type 'int', received {type(value)}."
            raise ValueError(msg)

        return value

    if issubclass(data_type, Real):
        if not isinstance(value, float):
            msg = f"Expected value to be of type 'float', received {type(value)}."
            raise ValueError(msg)

        return value

    if issubclass(data_type, Binary):
        if not isinstance(value, bytes):
            msg = f"Expected value to be of type 'bytes', received {type(value)}."
            raise ValueError(msg)

        return value

    if issubclass(data_type, Boolean):
        if not isinstance(value, bool):
            msg = f"Expected value to be of type 'bool', received {type(value)}."
            raise ValueError(msg)

        return value

    if issubclass(data_type, sila.Any):
        return value

    msg = f"Received unknown data type {data_type.__name__}."
    raise ValueError(msg)



import dataclasses
import json
import pathlib
import re
import unittest.mock

import pydantic
import pytest
import typing_extensions as typing
from ruamel.yaml import YAML

from unitelabs.cdk.config import (
    Config,
    ConfigurationError,
    ConnectorBaseConfig,
    Field,
    InvalidSchemaFieldError,
    UnsupportedConfigFiletype,
    delayed_default,
    read_config_file,
    validate_config,
    validate_field,
)


class TestGetPydanticCoreSchema:
    def test_should_merge_base_class_descriptions(self):
        @dataclasses.dataclass
        class BaseConfig(Config):
            static: str = "DEFAULT"
            """Static string."""
            modified: int = 0
            """The old value."""

        BASE_CONFIG_SCHEMA = {
            "type": "object",
            "title": "BaseConfig",
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "properties": {
                "modified": {
                    "type": "integer",
                    "default": 0,
                    "title": "Modified",
                    "description": "The old value.",
                },
                "static": {
                    "type": "string",
                    "default": "DEFAULT",
                    "title": "Static",
                    "description": "Static string.",
                },
            },
        }

        assert BaseConfig.schema() == BASE_CONFIG_SCHEMA

        @dataclasses.dataclass
        class DerivedConfig(BaseConfig):
            modified: int = 100
            """The new value."""

        derived_schema = BASE_CONFIG_SCHEMA.copy()
        derived_schema["title"] = "DerivedConfig"
        derived_schema["properties"]["modified"]["default"] = 100
        derived_schema["properties"]["modified"]["description"] = "The new value."
        assert DerivedConfig.schema() == derived_schema


class TestToPydanticDataclass:
    def test_should_not_alter_class(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        pydantic_dataclass = ExampleConfig.to_pydantic_dataclass()
        assert pydantic.dataclasses.is_pydantic_dataclass(pydantic_dataclass)
        assert not pydantic.dataclasses.is_pydantic_dataclass(ExampleConfig)

    def test_returned_pydantic_dataclass_creates_instances_of_class(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        pydantic_dataclass = ExampleConfig.to_pydantic_dataclass()
        instance = pydantic_dataclass()
        assert isinstance(instance, ExampleConfig)


class TestDelayedDefault:
    def test_should_call_set_value(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = dataclasses.field(default_factory=delayed_default(lambda self: 1))

        example = ExampleConfig()
        assert example.value == 1

    def test_should_allow_use_of_other_fields(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0
            derived_value: int = dataclasses.field(default_factory=delayed_default(lambda self: self.value + 1))

        example = ExampleConfig()
        assert example.value == 0
        assert example.derived_value == 1

    def test_should_call_method(self):
        method = unittest.mock.Mock()
        method.return_value = 1

        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0
            derived_value: int = dataclasses.field(default_factory=delayed_default(lambda self: method(self)))

        example = ExampleConfig()
        assert example.value == 0
        assert example.derived_value == 1
        method.assert_called_once_with(example)

    @pytest.mark.parametrize(
        "entrypoint,expand",
        [
            pytest.param("validate", False, id="validate"),
            pytest.param("from_dict", False, id="from_dict"),
            pytest.param("cls", True, id="from constructor"),
        ],
    )
    def test_should_allow_override(self, entrypoint, expand):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            type: typing.Literal["a", "b"] = "a"
            value: str = dataclasses.field(default_factory=delayed_default(lambda self: f"{self.type} value"))

        test_method = ExampleConfig if entrypoint == "cls" else getattr(ExampleConfig, entrypoint)

        # Test default value
        config = test_method(type="b") if expand else test_method({"type": "b"})
        assert config.type == "b"
        assert config.value == "b value"

        # Test custom override
        custom_value = "custom value"
        config = (
            test_method(type="a", value=custom_value) if expand else test_method({"type": "a", "value": custom_value})
        )
        assert config.type == "a"
        assert config.value == custom_value

    def test_should_allow_partial_override(self):
        @dataclasses.dataclass
        class NestedConfig(Config):
            type: typing.Literal["A", "B"] = "A"
            name: str = dataclasses.field(default_factory=delayed_default(lambda self: f"{self.type} name"))

        @dataclasses.dataclass
        class MyConfig(Config):
            nested: NestedConfig = dataclasses.field(default_factory=NestedConfig)
            nested_name: str = dataclasses.field(default_factory=delayed_default(lambda self: self.nested.name))

        config = MyConfig.validate({})
        assert config.nested.type == "A"
        assert config.nested_name == config.nested.name

        config = MyConfig.validate({"nested_name": "custom name"})
        assert config.nested.type == "A"
        assert config.nested.name == "A name"
        assert config.nested_name == "custom name"

        config = MyConfig.validate({"nested": {"type": "B"}})
        assert config.nested.type == "B"
        assert config.nested.name == config.nested_name == "B name"

    def test_should_catch_circular_usages(self):
        @dataclasses.dataclass
        class AConfig(Config):
            type: typing.Literal["A1", "A2"] = dataclasses.field(
                default_factory=delayed_default(lambda self: "A1" if "A" in self.name else "A2")
            )
            name: str = dataclasses.field(default_factory=delayed_default(lambda self: f"{self.type} name"))

        a = AConfig()
        match = (
            r"The delayed_default value for AConfig.(type|name) could not be resolved. "
            "This is most likely due to a circular dependency between delayed_default values."
        )  # does not consistently resolve to type or name
        with pytest.raises(ConfigurationError, match=match):
            a.type

        with pytest.raises(ConfigurationError, match=match):
            a.name


class TestFromDict:
    def test_should_allow_partial_set(self):
        @dataclasses.dataclass
        class ExampleConfig(ConnectorBaseConfig):
            a: int = 0
            b: int = 1

        config = ExampleConfig.from_dict({"a": 10})
        assert config
        assert config.a == 10
        assert config.b == 1

    def test_should_raise_ConfigurationError_on_invalid_field(self):
        @dataclasses.dataclass
        class ExampleConfig(ConnectorBaseConfig):
            value: int = 0

        with pytest.raises(
            ConfigurationError,
            match="Invalid configuration for ExampleConfig: 1 validation error for ExampleConfig"
            "\nvalue"
            "\n  Input should be a valid integer",
        ):
            ExampleConfig.from_dict({"value": "not an int"})

    def test_should_allow_arbitrary_fields(self):
        @dataclasses.dataclass
        class ExampleConfig(ConnectorBaseConfig):
            value: int = 0

        config = ExampleConfig.from_dict({"value": 0, "arbitrary_field": True})
        assert config
        assert config.value == 0
        assert config.arbitrary_field

    def test_should_not_raise_ConfigurationError_on_invalid_field_set(self):
        @dataclasses.dataclass
        class ExampleConfig(ConnectorBaseConfig):
            value: int = 0

        example = ExampleConfig.from_dict({"value": 0})
        example.value = "not an int"

        assert example.value == "not an int"

    def test_should_set_source(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        config = ExampleConfig.from_dict({"value": 0})
        assert config._source == {"value": 0}
        assert config._source_path is None


class TestToDict:
    def test_should_return_dict(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        default = ExampleConfig().to_dict()
        assert isinstance(default, dict)
        assert default == {"value": 0}

    def test_should_allow_arbitrary_fields(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        config = ExampleConfig.from_dict({"value": 0, "arbitrary_field": True})
        data = config.to_dict()
        assert isinstance(data, dict)
        assert data == {"value": 0, "arbitrary_field": True}


class TestDefault:
    def test_should_return_dict(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        default = ExampleConfig().to_dict()
        assert isinstance(default, dict)
        assert default == {"value": 0}

    def test_added_keys_included(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            a: int = 0
            b: float = 0.0
            c: str = ""
            d: bool = True
            e: list[str] = dataclasses.field(default_factory=list)
            f: dict[str, int] = dataclasses.field(default_factory=dict)

        assert ExampleConfig().to_dict() == {"a": 0, "b": 0.0, "c": "", "d": True, "e": [], "f": {}}


class TestDescribe:
    def test_should_return_dict(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0
            """The value."""

        default = ExampleConfig().describe()
        assert isinstance(default, dict)
        assert default == {"value": {"type": "int", "default": "0", "description": "The value."}}

    def test_should_return_dict_for_nested_field(self):
        @dataclasses.dataclass
        class NestedConfig(Config):
            value: int = 0
            """The value."""

        @dataclasses.dataclass
        class ExampleConfig(Config):
            nested: NestedConfig = dataclasses.field(default_factory=NestedConfig)
            """The nested config."""

        default = ExampleConfig().describe()
        assert isinstance(default, dict)
        FULL_DATA = {
            "nested": {
                "values": {"value": {"type": "int", "default": "0", "description": "The value."}},
                "description": "For more information use `config show --field nested`.",
                "type": "NestedConfig",
                "default": "",
            }
        }
        assert default == FULL_DATA
        assert ExampleConfig().describe("nested") == FULL_DATA["nested"]["values"]

    def test_should_raise_invalid_schema_field_error(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        with pytest.raises(
            InvalidSchemaFieldError,
            match="'non_existent_field' is not a valid field in ExampleConfig.",
        ):
            ExampleConfig.describe("non_existent_field")


class TestLoad:
    @pytest.mark.parametrize("ext", ["toml", "ini"])
    def test_should_raise_unsupported_filetype(self, tmp_path, ext):
        config_file_path = tmp_path / f"config.{ext}"
        config_file_path.touch()
        with pytest.raises(
            UnsupportedConfigFiletype,
            match=f"Cannot read file at '{config_file_path.resolve()}'. Only yaml and json filetypes are supported.",
        ):
            ConnectorBaseConfig.load(config_file_path)

    @pytest.mark.parametrize("ext", ["json", "yaml"])
    def test_should_set_source_and_source_path(self, tmp_path, ext):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        config_path = tmp_path / f"config.{ext}"
        config = ExampleConfig().dump(config_path)

        config = ExampleConfig.load(config_path)
        assert config._source == {"value": 0}
        assert config._source_path == config_path


class TestDump:
    def test_should_dump_to_json(self, tmp_path):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        config = ExampleConfig()
        config_path = tmp_path / "config.json"
        config.dump(config_path)

        with config_path.open("r") as f:
            data = json.load(f)

        assert data == {"value": 0}

    def test_should_dump_to_yaml(self, tmp_path):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        config = ExampleConfig()
        config_path = tmp_path / "config.yaml"
        config.dump(config_path)

        yaml = YAML(typ="safe")
        with config_path.open("r") as f:
            data = yaml.load(f)

        assert data == {"value": 0}

    def test_should_preserve_yaml_comments(self, tmp_path):
        sample_yaml = """# before comment\nvalue: 0 # comment\n"""

        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        in_config_path = tmp_path / "config.yaml"

        with in_config_path.open("w") as f:
            f.write(sample_yaml)

        config = ExampleConfig.load(in_config_path)
        assert config.value == 0

        out_config_path = tmp_path / "out_config.yaml"
        config.dump(out_config_path)

        with out_config_path.open("r") as f:
            data = f.read()
            assert data == sample_yaml

    def test_should_allow_arbitrary_fields(self, tmp_path):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        config = ExampleConfig.from_dict({"value": 0, "arbitrary_field": True})
        config_path = tmp_path / "config.json"
        config.dump(config_path)

        with config_path.open("r") as f:
            data = json.load(f)

        assert data == {"value": 0, "arbitrary_field": True}


class TestFieldValidation_Basic:
    def test_should_format_configuration_errors_raised_in_post_init(self):
        msg = "value must be greater than 1."

        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

            def __post_init__(self):
                if not self.value > 1:
                    raise ConfigurationError(msg)

        with pytest.raises(
            ConfigurationError,
            match=(
                f"Invalid configuration for ExampleConfig: 1 validation error for ExampleConfig\n  Value error, {msg}"
            ),
        ):
            ExampleConfig.validate({"value": 0})

    def test_should_not_raise_ValidationError_from_post_init_validation_on_field_set(self):
        msg = "value must be greater than 1."

        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

            def __post_init__(self):
                if not self.value > 0:
                    raise ConfigurationError(msg)

        config = ExampleConfig.validate({"value": 2})
        assert config.value == 2

        config.value = 0

    def test_should_validate_types(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        with pytest.raises(
            ConfigurationError,
            match="Invalid configuration for ExampleConfig: "
            "1 validation error for ExampleConfig\n"
            "value\n  "
            "Input should be a valid integer, unable to parse string as an integer",
        ):
            ExampleConfig.validate({"value": "string"})

    def test_should_raise_ValidationError_for_invalid_type_on_field_set(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: int = 0

        config = ExampleConfig.validate({"value": 2})
        assert config.value == 2

        with pytest.raises(
            pydantic.ValidationError,
            match="1 validation error for ExampleConfig\nvalue\n  Input should be a valid integer",
        ):
            config.value = "string"

    def test_should_raise_for_invalid_field_name(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            value: bool = False

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "Provided field or fields ['invalid_field'] are not valid for ExampleConfig configuration."
            ),
        ):
            ExampleConfig.validate({"invalid_field": True, "value": True})


class TestFieldValidation_FieldAnnotation:
    def test_should_validate_with_field_annotation(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            large_int: typing.Annotated[int, Field(gt=9000)] = 9001

        default = ExampleConfig().to_dict()
        assert ExampleConfig.validate(default)

        default.update({"large_int": 29})
        with pytest.raises(
            ConfigurationError,
            match="Invalid configuration for ExampleConfig: 1 validation error for ExampleConfig"
            "\nlarge_int"
            "\n  Input should be greater than 9000",
        ):
            ExampleConfig.validate(default)

    def test_should_raise_ValidationError_on_invalid_field_set(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            large_int: typing.Annotated[int, Field(gt=9000)] = 9001

        config = ExampleConfig.validate({"large_int": 9500})
        assert config.large_int == 9500

        with pytest.raises(
            pydantic.ValidationError,
            match="1 validation error for ExampleConfig\nlarge_int\n  Input should be greater than 9000",
        ):
            config.large_int = 29


class TestFieldValidation_ValidateFieldDecorator:
    def test_should_validate_with_field_decorator(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            simple: bool = False

            @validate_field("simple")
            @classmethod
            def must_be_true(cls, value: bool) -> bool:
                if not value:
                    msg = "simple must be True."
                    raise ConfigurationError(msg)
                return value

        default = ExampleConfig().to_dict()
        with pytest.raises(
            ConfigurationError,
            match="Invalid configuration for ExampleConfig: 1 validation error for ExampleConfig"
            "\nsimple"
            "\n  Value error, simple must be True.",
        ):
            ExampleConfig.validate(default)

        default.update({"simple": True})
        example = ExampleConfig.validate(default)
        assert example.simple

    def test_should_raise_ValidationError_on_invalid_field_set(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            simple: bool = False

            @validate_field("simple")
            @classmethod
            def must_be_true(cls, value: bool) -> bool:
                if not value:
                    msg = "simple must be True."
                    raise ConfigurationError(msg)
                return value

        example = ExampleConfig.validate({"simple": True})
        with pytest.raises(pydantic.ValidationError):
            example.simple = False


class TestConfigValidation_ValidateConfigDecorator:
    def test_should_validate_with_validate_config_decorator(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            complex: bool = False
            """Whether or not the complex configuration is enabled."""
            complex_name: typing.Optional[str] = ""
            """The name of the complex configuration."""

            @validate_config()
            def must_be_named(self) -> typing.Self:
                if self.complex and not self.complex_name:
                    msg = "complex configuration requires additional 'complex_name' value."
                    raise ConfigurationError(msg)
                return self

        default = ExampleConfig().to_dict()
        config = ExampleConfig.validate(default)

        assert hasattr(config, "complex")
        assert hasattr(config, "complex_name")

        with pytest.raises(
            ConfigurationError,
            match="Invalid configuration for ExampleConfig: 1 validation error for ExampleConfig"
            "\n  Value error, complex configuration requires additional 'complex_name' value.",
        ):
            default.update({"complex": True})
            ExampleConfig.validate(default)

    def test_should_validate_with_mulitiple_validate_config_decorators(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            complex: bool = False
            complex_name: typing.Optional[str] = ""

            @validate_config()
            def must_be_named(self) -> typing.Self:
                if self.complex and not self.complex_name:
                    msg = "complex configuration requires additional 'complex_name' value."
                    raise ConfigurationError(msg)
                return self

            @validate_config()
            def must_contain_substring(self) -> typing.Self:
                if (self.complex and self.complex_name) and not self.complex_name.startswith("A"):
                    msg = "'complex_name' must start with 'A'."
                    raise ConfigurationError(msg)
                return self

        default = ExampleConfig().to_dict()
        config = ExampleConfig.validate(default)

        assert hasattr(config, "complex")
        assert hasattr(config, "complex_name")

        with pytest.raises(
            ConfigurationError,
            match="Invalid configuration for ExampleConfig: 1 validation error for ExampleConfig"
            "\n  Value error, 'complex_name' must start with 'A'.",
        ):
            default.update({"complex": True, "complex_name": "this"})
            ExampleConfig.validate(default)

    def test_should_raise_ValidationError_on_invalid_field_set(self):
        @dataclasses.dataclass
        class ExampleConfig(Config):
            complex: bool = False
            complex_name: typing.Optional[str] = ""

            @validate_config()
            def must_be_named(self) -> typing.Self:
                if self.complex and not self.complex_name:
                    msg = "complex configuration requires additional 'complex_name' value."
                    raise ConfigurationError(msg)
                return self

        config = ExampleConfig.validate({"complex": False, "complex_name": ""})
        with pytest.raises(
            pydantic.ValidationError,
            match="1 validation error for ExampleConfig\n  Value error, complex configuration requires additional 'complex_name' value.",
        ):
            config.complex = True


class TestReadConfigFile:
    def test_should_raise_if_default_not_found(self):
        with pytest.raises(
            FileNotFoundError, match="No config was provided and none was found at any of the default paths:"
        ):
            read_config_file()

    def test_should_raise_if_file_not_found(self):
        path = pathlib.Path("config.json")
        with pytest.raises(FileNotFoundError, match=f"File at path '{path.resolve()}' not found."):
            read_config_file(path)

    @pytest.mark.parametrize("ext", ["json", "yaml"])
    def test_should_read_all_data(self, tmp_path, ext):
        config_path = tmp_path / f"config.{ext}"

        @dataclasses.dataclass
        class ExampleConfig(Config):
            name: typing.Annotated[str, Field(min_length=1)] = "Default"
            """The name of the example."""
            must_be_set: typing.Annotated[int, Field(ge=0, le=10)] = -1
            """A value with an invalid default."""

        config = ExampleConfig(name="Example", must_be_set=1)
        config.dump(config_path)

        reloaded = read_config_file(config_path)
        assert reloaded == config.to_dict()



from .simulation_controller_base import SimulationControllerBase, StartSimulationModeFailed, StartRealModeFailed

__all__ = ["SimulationControllerBase", "StartRealModeFailed", "StartSimulationModeFailed"]



import unittest.mock

import pytest
from sila.server import discovery

from unitelabs.cdk.connector import Connector


@pytest.fixture
def no_broadcast(monkeypatch):
    class FakeBroadcaster(discovery.Discovery):
        """Fake Broadcaster."""

        def __init__(self, *args, **kwargs):
            self.start = unittest.mock.AsyncMock()
            self.stop = unittest.mock.AsyncMock()

    monkeypatch.setattr(discovery, "Discovery", FakeBroadcaster)


class TestConnector:
    # Create a new connector with default values
    async def test_create_new_connector_with_default_values(self, no_broadcast):
        connector = Connector()

        assert connector.config.sila_server.hostname == "0.0.0.0"
        assert connector.config.sila_server.port == 0
        assert not connector.config.sila_server.tls
        assert not connector.config.sila_server.require_client_auth
        assert not connector.config.sila_server.root_certificates
        assert not connector.config.sila_server.certificate_chain
        assert not connector.config.sila_server.private_key
        assert connector.config.sila_server.version == "0.1"
        assert connector.config.sila_server.name == "SiLA Server"
        assert connector.config.sila_server.type == "ExampleServer"
        assert connector.config.sila_server.vendor_url == "https://sila-standard.com"

        assert connector.config.cloud_server_endpoint.hostname == "localhost"
        assert connector.config.cloud_server_endpoint.port == 50000
        assert connector.config.cloud_server_endpoint.tls
        assert not connector.config.cloud_server_endpoint.root_certificates
        assert not connector.config.cloud_server_endpoint.certificate_chain
        assert not connector.config.cloud_server_endpoint.private_key
        assert connector.config.cloud_server_endpoint.reconnect_delay == 10000
        assert connector.config.cloud_server_endpoint.options == {
            "grpc.keepalive_time_ms": 60 * 1000,
            "grpc.keepalive_timeout_ms": 3 * 60 * 1000,
            "grpc.keepalive_permit_without_calls": 0,
            "grpc.http2.max_pings_without_data": 0,
        }


class TestStart:
    # The start method calls stop after cancellation
    async def test_start_calls_stop_on_cancellation(self, no_broadcast):
        with (
            unittest.mock.patch("sila.server.Server", spec=True),
            unittest.mock.patch("sila.server.Discovery", spec=True),
            unittest.mock.patch("sila.server.CloudServer", spec=True),
        ):
            connector = Connector()
            connector.stop = unittest.mock.AsyncMock()

            await connector.start()
            await connector.stop()

            connector.stop.assert_awaited_once_with()


class TestClose:
    # The stop method calls shutdown handlers
    async def test_stop_calls_shutdown_handlers(self, no_broadcast):
        with (
            unittest.mock.patch("sila.server.Server", spec=True),
            unittest.mock.patch("sila.server.Discovery", spec=True),
            unittest.mock.patch("sila.server.CloudServer", spec=True),
        ):
            handler = unittest.mock.Mock()
            connector = Connector()
            connector.on_shutdown(handler=handler)

            await connector.stop()

            handler.assert_called_once_with()

    # The stop method calls shutdown handlers in order of registration
    async def test_stop_calls_shutdown_handlers_in_order(self, no_broadcast):
        with (
            unittest.mock.patch("sila.server.Server", spec=True),
            unittest.mock.patch("sila.server.Discovery", spec=True),
            unittest.mock.patch("sila.server.CloudServer", spec=True),
        ):
            handler = unittest.mock.Mock()
            handler.handler_1 = unittest.mock.Mock()
            handler.handler_2 = unittest.mock.Mock()
            connector = Connector()
            connector.on_shutdown(handler=handler.handler_1)
            connector.on_shutdown(handler=handler.handler_2)

            await connector.stop()

            handler.assert_has_calls([unittest.mock.call.handler_1(), unittest.mock.call.handler_2()])

    # The stop method calls async shutdown handlers
    async def test_stop_calls_async_shutdown(self, no_broadcast):
        with (
            unittest.mock.patch("sila.server.Server", spec=True),
            unittest.mock.patch("sila.server.Discovery", spec=True),
            unittest.mock.patch("sila.server.CloudServer", spec=True),
        ):
            handler = unittest.mock.AsyncMock()
            connector = Connector()
            connector.on_shutdown(handler=handler)

            await connector.start()
            await connector.stop()

            handler.assert_awaited_once_with()

    # The stop method ignores exception in shutdown handlers
    async def test_stop_ignores_exception_in_shutdown(self, no_broadcast):
        with (
            unittest.mock.patch("sila.server.Server", spec=True),
            unittest.mock.patch("sila.server.Discovery", spec=True),
            unittest.mock.patch("sila.server.CloudServer", spec=True),
        ):
            handler_1 = unittest.mock.Mock(side_effect=RuntimeError)
            handler_2 = unittest.mock.AsyncMock(side_effect=RuntimeError)
            connector = Connector()
            connector.on_shutdown(handler=handler_1)
            connector.on_shutdown(handler=handler_2)

            await connector.start()
            await connector.stop()

            handler_1.assert_called_once_with()
            handler_2.assert_awaited_once_with()


class TestOnShutdown:
    # Can add a shutdown handler to the list of handlers
    async def test_add_shutdown_handler(self, no_broadcast):
        connector = Connector()

        def shutdown_handler():
            print("Shutdown handler called")

        connector.on_shutdown(shutdown_handler)

        assert len(connector._shutdown_handlers) == 1
        assert connector._shutdown_handlers[0] == shutdown_handler

    # Adding a shutdown handler with a non-callable object raises a TypeError
    async def test_add_non_callable_handler_raises_type_error(self, no_broadcast):
        connector = Connector()
        non_callable_handler = "not a callable object"

        with pytest.raises(TypeError, match=r"The `handler` argument must be callable."):
            connector.on_shutdown(
                non_callable_handler,  # type: ignore
            )


class TestOffShutdown:
    # Can remove a previously added shutdown hook
    async def test_remove_shutdown_hook(self, no_broadcast):
        # Initialize the class object
        connector = Connector()

        # Define a mock handler
        mock_handler = unittest.mock.Mock()

        # Add the mock handler to the shutdown handlers list
        connector.on_shutdown(mock_handler)

        # Remove the mock handler using off_shutdown method
        connector.off_shutdown(mock_handler)

        # Assert that the mock handler is no longer in the shutdown handlers list
        assert mock_handler not in connector._shutdown_handlers

    # removing a shutdown hook from an empty list of shutdown handlers
    async def test_remove_shutdown_hook_from_empty_list(self, no_broadcast):
        # Initialize the class object
        connector = Connector()

        # Define a mock handler
        mock_handler = unittest.mock.Mock()

        # Remove the mock handler using off_shutdown method
        connector.off_shutdown(mock_handler)

        # Assert that the mock handler is still not in the shutdown handlers list
        assert mock_handler not in connector._shutdown_handlers



import dataclasses
import functools
import json
import pathlib

import pydantic
import pydantic_core
import typing_extensions as typing
from pydantic import Field, model_validator
from pydantic import field_validator as validate_field
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue
from ruamel.yaml import YAML, CommentedMap

validate_config = functools.partial(model_validator, mode="after")


BasicSerializableType: typing.TypeAlias = typing.Union[str, int, float, bool]
SerializableType: typing.TypeAlias = typing.Union[
    BasicSerializableType,
    None,
    list["SerializableType"],
    dict[str, "SerializableType"],
]
SerializableDict: typing.TypeAlias = dict[str, SerializableType]
DEFAULT_CONFIG_PATHS = [pathlib.Path("./config.json"), pathlib.Path("./config.yaml"), pathlib.Path("./config.yml")]


class UnsupportedConfigFiletype(Exception):
    """The filetype is unsupported for reading/writing config files."""


class ConfigurationError(ValueError):
    """Received an invalid configuration."""


def get_schema_fields(data: pydantic_core.core_schema.CoreSchema) -> list[SerializableDict]:
    """Get the fields from a pydantic core schema."""
    schema = data.get("schema", {})
    if (fields := schema.get("fields")) and isinstance(fields[0], dict):
        return fields
    return get_schema_fields(schema.get("schema", {})) if schema else []


class JsonSchemaGenerator(GenerateJsonSchema):
    """
    Custom JSON Schema generator for compliance with UniteLabs PEP-17: JSON Schema.

    More info about the specification can be found at: https://www.notion.so/unitelabs/JSON-Schema-13403b686b5f8099910cf52b9e1510b5
    """

    def generate(self, schema: pydantic_core.core_schema.CoreSchema, mode: str = "validation") -> JsonSchemaValue:
        json_schema = super().generate(schema, mode)
        json_schema["$schema"] = "https://json-schema.org/draft/2020-12/schema"
        return json_schema

    def bytes_schema(self, schema: pydantic_core.core_schema.BytesSchema) -> JsonSchemaValue:
        byte_schema = super().bytes_schema(schema)
        byte_schema["contentEncoding"] = "base64"
        return byte_schema


def update_yaml(yaml: CommentedMap, data: dict) -> None:
    """
    Recursively update a `ruamel.yaml.CommentedMap` with data from a dictionary.

    Args:
      yaml: The `CommentedMap` to update.
      data: A dictionary with keys matching those contained in `yaml` from which updated values
       will be applied to the `CommentedMap`.
    """
    for key, value in data.items():
        if isinstance(value, dict):
            update_yaml(yaml[key], value)
        else:
            current = yaml[key]
            if current != value:
                yaml[key] = value


class MissingDefault:
    def __repr__(self):
        return "MissingDefault"


UNCONFIGURED = MissingDefault()
T = typing.TypeVar("T")


class DelayedDefault(typing.Generic[T]):
    def __init__(self, func: typing.Callable[["Config"], T]):
        self.func = func
        self.resolved = False
        self.value = typing.cast(T, UNCONFIGURED)

    def resolve(self, instance: "Config") -> T:
        if not self.resolved:
            self.value = self.func(instance)
            self.resolved = True
        return self.value


def delayed_default(func: typing.Callable[["Config"], T]) -> typing.Callable[[], T]:
    def factory() -> T:
        return typing.cast(T, DelayedDefault(func))

    return factory


class Config:
    """A pydantic-enabled dataclass that represents a configuration."""

    __pydantic_config__ = pydantic.ConfigDict(
        validate_assignment=True,
        revalidate_instances="always",
        use_attribute_docstrings=True,
        ser_json_bytes="base64",
        val_json_bytes="base64",
    )
    _ignore: typing.ClassVar[set[str]] = {"_source", "_source_path"}
    _source: typing.Union[CommentedMap, SerializableDict, None] = None
    _source_path: typing.Optional[pathlib.Path] = None

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: typing.Self, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.core_schema.CoreSchema:
        derived_class = handler(cls)

        base = [c for c in cls.__mro__ if dataclasses.is_dataclass(c) and c not in (cls, Config)]

        if not base or len(base) != 1 or len(dataclasses.fields(base[0])) < 1:
            return derived_class

        base_class = handler(base[0])

        base_fields_by_name = {field["name"]: field for field in get_schema_fields(base_class)}

        for derived_field in get_schema_fields(derived_class):
            base_field = base_fields_by_name.get(derived_field["name"], {})
            if not derived_field.get("metadata") and (update := base_field.get("metadata")):
                derived_field["metadata"] = update

        return derived_class

    def __getattribute__(self, name: str):
        value = super().__getattribute__(name)

        if isinstance(value, DelayedDefault):
            try:
                return value.resolve(self)
            except RecursionError:
                msg = (
                    f"The delayed_default value for {self.__class__.__name__}.{name} could not be resolved. "
                    "This is most likely due to a circular dependency between delayed_default values."
                )
                raise ConfigurationError(msg) from None
        return value

    @classmethod
    def to_pydantic_dataclass(cls) -> type[typing.Self]:
        """Create a pydantic dataclass from the Config."""
        subclass = type(cls.__name__, (cls,), {})
        return pydantic.dataclasses.dataclass(subclass)

    @classmethod
    def get_type_adapter(cls) -> pydantic.TypeAdapter:
        """Get a pydantic type adapter for this class."""
        return pydantic.TypeAdapter(cls)

    @classmethod
    def from_dict(cls, data: SerializableDict) -> typing.Self:
        """
        Create a new `Config` instance from a dict, without rejecting unknown fields.

        This function allows arbitrary additional data to be set on the `Config` object
        and therefore does not guarantee validations on setting attributes for all fields, unlike `validate`.

        In this way, this method should only be used in contexts where the derived configuration is not present,
        e.g. `certificate` CLI where app is not loaded and thus the subclass of `ConnectorBaseConfig` cannot be found.

        Args:
          data: A dictionary of configuration values.

        Returns:
          A validated `Config` instance, where only the known fields have been validated.
        """
        fields = [field.name for field in dataclasses.fields(cls)]
        known_fields = {k: v for k, v in data.items() if k in fields}
        unknown_fields = {k: v for k, v in data.items() if k not in known_fields}

        cls.validate(known_fields)

        final = cls(**known_fields)
        for k, v in unknown_fields.items():
            setattr(final, k, v)

        final._source = data

        return final

    def to_dict(self) -> SerializableDict:
        """Get the serializable dictionary representation of the instance."""
        data = json.loads(self.get_type_adapter().dump_json(self))
        if len(dataclasses.fields(self)) < len(self.__dict__.keys()):
            # enables dump when a config has been loaded from_dict
            known_fields = [f.name for f in dataclasses.fields(self)]
            additional_data = {k: v for k, v in self.__dict__.items() if k not in [*known_fields, *self._ignore]}
            for key, value in additional_data.items():
                data[key] = value
        return data

    @classmethod
    def schema(cls) -> dict[str, typing.Any]:
        """Get the JSON Schema for this class."""
        return cls.get_type_adapter().json_schema(schema_generator=JsonSchemaGenerator)

    @classmethod
    def describe(cls, field: typing.Optional[str] = None) -> dict[str, typing.Any]:
        """
        Get a description of the whole configuration or for a field in the configuration.

        Currently only supports fields that are present in the schema at depth 1.

        Args:
          field: The field in the `Config` to get a description for.

        Raises:
          InvalidSchemaFieldError: If the provided `field` is not present in the schema.
        """
        from .schema import InvalidSchemaFieldError, Schema, describe

        schema = Schema(cls.schema())
        description = describe(cls, schema)
        if field is None:
            return description

        if field not in description:
            msg = f"'{field}' is not a valid field in {cls.__name__}."
            raise InvalidSchemaFieldError(msg)
        return description[field].get("values", description[field])

    @classmethod
    def load(cls, path: typing.Optional[pathlib.Path] = None, strict: bool = False) -> typing.Self:
        """
        Load a connector configuration from `path`.

        If no `path` is provided, searches the default config file locations: `./config.json`, `./config.yaml`,
        and `./config.yml` for an existing config file.

        Args:
          path: The path to the configuration file, can be a yaml or json filetype,
            defaults to first found config file in default locations.
          strict: Whether or not to raise an error if the file contains fields not defined in the `Config` dataclass.

        Returns:
          A `Config` instance, where only the known fields have been validated if not `strict`,
            or all fields have been validated if `strict`.

        Raises:
          FileNotFoundError: If no config file is found at the provided `path`.
          UnsupportedConfigFiletype: If the provided `path` is not a yaml or json file.
          ConfigurationError: If the config file contains invalid values.
        """
        path = find_file(path, default_paths=DEFAULT_CONFIG_PATHS)  # call find_file explicitly to resolve source path
        data = read_config_file(path)

        inst: typing.Self
        inst = cls.validate(data) if strict else cls.from_dict(data)
        inst._source = data
        inst._source_path = path

        return inst

    def dump(self, path: pathlib.Path) -> None:
        """
        Write the current configuration to a file.

        Args:
          path: The path at which to write the configuration, may be yaml or json filetype.

        Raises:
          UnsupportedConfigFiletype: If `path` extension is not `.yaml` or `.json`.
        """
        data = self.to_dict()

        if path.suffix == ".json":
            data = json.dumps(data, indent=2)
            with path.open("w") as f:
                f.write(data)
            return
        if path.suffix in [".yaml", ".yml"]:
            yaml = YAML()
            if hasattr(self, "_source_path") and self._source_path and self._source_path.exists():
                # preserve width of source file
                yaml.width = max(map(len, self._source_path.read_text().splitlines()))
            with path.open("w") as f:
                if hasattr(self, "_source") and isinstance(self._source, CommentedMap):
                    update_yaml(self._source, data)
                    data = self._source
                yaml.dump(data, f)

            return
        msg = f"Cannot write file to {path}. Only yaml and json filetypes are supported."
        raise UnsupportedConfigFiletype(msg)

    @classmethod
    def validate(cls, values: typing.Optional[SerializableDict]) -> typing.Self:
        """
        Validate the configuration values.

        Args:
          values: The configuration values to validate.

        Returns:
          A validated `Config` instance, or the default instance if no values are provided.
        """

        values = values if values is not None else {}

        if unknown_fields := [k for k in values if k not in [f.name for f in dataclasses.fields(cls)]]:
            msg = f"Provided field or fields {unknown_fields} are not valid for {cls.__name__} configuration."
            raise ConfigurationError(msg)

        try:
            configuration = cls.to_pydantic_dataclass()(**values)
        except pydantic.ValidationError as e:
            msg = f"Invalid configuration for {cls.__name__}: {e}"
            raise ConfigurationError(msg) from None

        return configuration


def read_config_file(path: typing.Optional[pathlib.Path] = None) -> SerializableDict:
    """
    Read in configuration data from a file.

    Args:
      path: The path to the configuration file, can be a yaml or json file,
        default checks `./config.json`, `./config.yaml`, `./config.yml` paths.

    Returns:
      A serializable dictionary of the configuration data.

    Raises:
      FileNotFoundError: If no file is found at the provided `path`.
      UnsupportedConfigFiletype: If the provided `path` is not a yaml or json file.
    """

    data = {}
    path = find_file(path, default_paths=DEFAULT_CONFIG_PATHS)

    if path.suffix == ".json":
        data = json.loads(path.read_text())
    elif path.suffix in (".yaml", ".yml"):
        yaml = YAML()
        data = yaml.load(path)
    else:
        msg = f"Cannot read file at '{path}'. Only yaml and json filetypes are supported."
        raise UnsupportedConfigFiletype(msg)

    return data


def find_file(
    path: typing.Optional[pathlib.Path] = None,
    default_paths: typing.Optional[list[pathlib.Path]] = None,
) -> pathlib.Path:
    """
    Search for a file at the provided path or in default locations.

    Args:
      path: The path to the file, can be a yaml or json file.
      default_paths: A list of default paths to search if no `path` is provided.

    Returns:
      The absolute path to the file.

    Raises:
      FileNotFoundError: If no file is found at the provided `path` or any of the `default_paths`.
    """

    if path is None:
        default_paths = default_paths or []
        path = get_extant_path(default_paths)
        if not path:
            resolved_paths = [p.resolve() for p in default_paths]
            msg = (
                f"No config was provided and none was found at any of the default paths: {resolved_paths}. "
                "Use the `config create` CLI to create a new configuration file."
            )
            raise FileNotFoundError(msg)

    path = path.resolve()
    if not path.exists():
        msg = f"File at path '{path}' not found."
        raise FileNotFoundError(msg)

    return path


def get_extant_path(paths: list[pathlib.Path]) -> typing.Optional[pathlib.Path]:
    """
    Search for an existing file from a list of paths.

    Args:
      paths: A list of paths, ordered by preference.

    Returns:
      The first absolute path from `paths` that exists or None, if none of the paths exist.
    """
    return next((path.resolve(strict=True) for path in paths if path.resolve().exists()), None)


__all__ = ["Config", "ConfigurationError", "Field", "UnsupportedConfigFiletype", "validate_config", "validate_field"]



import abc
import dataclasses
import typing

from unitelabs.cdk import sila


@dataclasses.dataclass
class PositionIndex(sila.CustomDataType):
    """Specifies a position via an index number, starting at 1."""

    position_index: typing.Annotated[int, sila.constraints.MinimalInclusive(value=1)]


@dataclasses.dataclass
class TargetPosition(sila.CustomDataType):
    """
    Represent a possible position of a device where the device can move to.

    Can contain a sub-position, e.g. for specifying a position in a rack.

    .. parameter:: The name of the target position (must be unique within the device).
    .. parameter:: The index of a sub-position within a target position or the number of sub-positions respectively,
                   e.g. for a rack.
    """

    position: str

    sub_position: PositionIndex


class MovementControllerBase(sila.Feature, metaclass=abc.ABCMeta):
    """
    This Feature provides control over the movement of devices.

    It specifies a set of predefined positions the device can be moved to, e.g. a robot arm can be moved to positioned
    taught earlier to the robot.
    """

    def __init__(self):
        super().__init__(
            originator="io.unitelabs",
            category="robot",
            version="1.0",
            maturity_level="Draft",
        )

    @abc.abstractmethod
    @sila.UnobservableProperty()
    async def get_available_positions(self) -> list[TargetPosition]:
        """Get all positions of the device including the number of sub-positions."""

    @abc.abstractmethod
    @sila.UnobservableProperty()
    async def get_current_position(self) -> TargetPosition:
        """Get the current position of the device."""

    @abc.abstractmethod
    @sila.ObservableCommand()
    async def move_to(
        self,
        target_position: TargetPosition,
        *,
        status: sila.Status,
    ) -> None:
        """
        Move the device to the specified position.

        .. parameter:: Indicates the position where the device will be moved to.
        """



import abc

from unitelabs.cdk import sila


class Unstable(Exception):
    """Command understood but timeout for stable reading was reached."""


class Overloaded(Exception):
    """Device in overload range."""


class Underloaded(Exception):
    """Device in underload range."""


class WeighingServiceBase(sila.Feature, metaclass=abc.ABCMeta):
    """
    This feature contains commands and properties used for common functions required when weighing things.

    The feature enables access to the current net weight (stable and dynamic) and the tare weight. Commands for zeroing
    and taring are provided.
    """

    def __init__(self, *args, **kwarg):
        super().__init__(
            originator="io.unitelabs",
            category="weighing",
            version="1.0",
            maturity_level="Draft",
            *args,  # noqa: B026
            **kwarg,
        )

    @abc.abstractmethod
    @sila.ObservableProperty(errors=[Overloaded, Underloaded])
    async def subscribe_weight(self) -> sila.Stream[float]:
        """Subscribe to the current net weight in gram, accessed immediately."""

    @abc.abstractmethod
    @sila.ObservableProperty()
    async def subscribe_tare_weight(self) -> sila.Stream[float]:
        """Subscribe to the stored tare weight in gram."""

    @abc.abstractmethod
    @sila.UnobservableCommand(errors=[Unstable, Overloaded, Underloaded])
    @sila.Response(name="Weight")
    def get_stable_weight(self) -> float:
        """
        Get the stable net weight in gram.

        .. return:: The stable net weight in gram.
        """

    @abc.abstractmethod
    @sila.UnobservableCommand()
    @sila.Response(name="Tare Weight")
    def tare(self) -> float:
        """
        Tare with the current net weight, executed immediately (Not stable).

        .. return:: The stored tare weight in gram.
        """

    @abc.abstractmethod
    @sila.UnobservableCommand(errors=[Unstable])
    @sila.Response(name="Tare Weight")
    def tare_stable(self) -> float:
        """
        Tare with the stable net weight.

        .. return:: The stored tare weight in gram.
        """

    @abc.abstractmethod
    @sila.UnobservableCommand()
    def set_tare_weight(self, tare_weight: float) -> None:
        """
        Set a new, custom tare weight in gram.

        .. parameter:: The tare weight to be stored in gram.
        """

    @abc.abstractmethod
    @sila.UnobservableCommand()
    def clear_tare_weight(self) -> None:
        """Clear the currently stored tare weight."""

    @abc.abstractmethod
    @sila.UnobservableCommand()
    def zero(self) -> None:
        """Zero the balance immediately."""

    @abc.abstractmethod
    @sila.UnobservableCommand(errors=[Unstable])
    def zero_stable(self) -> None:
        """Zero the balance with a stable measurement."""



import pathlib
import typing

import click

import sila
from unitelabs.cdk.config import get_connector_config


class MutuallyExclusiveOptions(Exception):
    """Two mutually exclusive options were used together."""


@click.group()
def certificate() -> None:
    """Handle certificates for TLS encryption."""


@certificate.command()
@click.option(
    "-cfg",
    "--config-path",
    type=click.Path(path_type=pathlib.Path, exists=True),
    default=None,
    required=False,
    help=(
        "The path to the configuration file which contains the UUID and host for the generated certificate, "
        "defaults to ./config.json if uuid and host not provided."
    ),
)
@click.option(
    "--uuid",
    type=str,
    required=False,
    help="The SiLA server's UUID.",
)
@click.option(
    "--host",
    type=str,
    required=False,
    help="The SiLA server's host address.",
)
@click.option(
    "--target",
    "-t",
    type=str,
    help="The output directory in which to store the certificate files.",
)
@click.option(
    "--embed",
    "-e",
    type=bool,
    is_flag=True,
    default=False,
    help=(
        "Whether or not to embed the certificate and key into the config file. "
        "Mutually exclusive with usage of '--target'."
    ),
)
@click.option(
    "--non-interactive",
    "-y",
    type=bool,
    is_flag=True,
    default=False,
    help=(
        "When using the `--config-path` option: "
        "suppress the input prompt and automatically update the config file's TLS values."
    ),
)
def generate(
    config_path: typing.Optional[pathlib.Path],
    uuid: typing.Optional[str],
    host: typing.Optional[str],
    target: typing.Optional[str],
    embed: bool,
    non_interactive: bool,
) -> None:
    """
    Generate a new self-signed certificate according to the SiLA 2 specification.

    Create a certificate with provided UUID and host.
    ```certificate generate --uuid <UUID> --host <host name>```

    Create a certificate using the UUID and host from the provided config file.
    ```certificate generate --config-path <path to config>```

    Create a certificate from a config file and update the config to enable TLS encryption.
    ```certificate generate --config-path <path to config> -y```

    The option `--config-path` cannot be used in combination with either `--host` or `--uuid`.

    If no `config-path`, `uuid`, or `host` is provided, searches the default config file locations:
    `./config.json`, `./config.yaml`, and `./config.yml` for an existing config file.
    """
    config_cls = get_connector_config()

    if target and embed:
        msg = "The option '--target' cannot be used with '--embed'."
        raise MutuallyExclusiveOptions(msg)

    if config_path and (uuid or host):
        msg = "The option '--config-path' cannot be used with '--uuid' or '--host'."
        raise MutuallyExclusiveOptions(msg)

    if not (uuid or host) and (config_path or config_path is None):
        config = config_cls.load(config_path)
        config_path = config_path or config._source_path

        uuid = config.sila_server.uuid
        host = config.sila_server.hostname

    key, cert = sila.server.generate_certificate(uuid, host)

    if not embed:
        target = target or "."
        click.echo(f"Writing certificate and key to directory: {target}")
        directory = pathlib.Path(target)
        directory.mkdir(parents=True, exist_ok=True)

        cert_path = directory / "cert.pem"
        cert_path.write_bytes(cert)
        cert = str(cert_path.resolve())

        key_path = directory / "key.pem"
        key_path.write_bytes(key)
        key = str(key_path.resolve())

    else:
        key = key.decode("ascii")
        cert = cert.decode("ascii")

    if config_path:
        if not non_interactive:
            response = click.prompt("Do you want to update your configuration file to enable TLS encryption? Enter Y/N")
            if response.lower() not in ["y", "yes"]:
                return

        config.sila_server.tls = True
        config.sila_server.private_key = key
        config.sila_server.certificate_chain = cert
        click.echo(f"Updating configuration file at: {config_path}.")
        config.dump(config_path)


if __name__ == "__main__":
    certificate()



import contextlib
import types
import weakref

import typing_extensions as typing

from .default import _DEFAULT_VALUE, Default
from .subscription import Subscription

IN = typing.TypeVar("IN")
OUT = typing.TypeVar("OUT", default=IN)
T = typing.TypeVar("T")


PipeFunction = typing.Callable[[IN], OUT]


class Subject(typing.Generic[IN, OUT], contextlib.AbstractContextManager):
    """
    An observable that can be updated externally and subscribed to by multiple observers.

    Args:
      maxsize: The maximum number of messages to track in `Subscription` queues created by `subscribe`.
    """

    _pipe: PipeFunction[IN, OUT]

    def __init__(self, maxsize: int = 0, pipe: typing.Optional[PipeFunction[IN, OUT]] = None) -> None:
        self._maxsize = maxsize
        self._value: typing.Union[OUT, Default] = _DEFAULT_VALUE

        self._total_subscribers = 0
        self._subscribers: list["Subscription[OUT]"] = []
        self._children: list[Subject[OUT, typing.Any]] = []
        self._parent: typing.Optional["Subject"] = None
        self._is_temporary = False
        self._context: typing.Optional["Subscription[OUT]"] = None

        def default_pipe(x: IN) -> OUT:
            return typing.cast(OUT, x)

        self._pipe = pipe or default_pipe

    def __repr__(self) -> str:
        name = self.__class__.__name__
        identifier = str(id(self))
        pipe = self._pipe.__name__ if hasattr(self._pipe, "__name__") else self._pipe
        n_subs, n_children = len(self.subscribers), len(self._children)
        return f"{name}({identifier}, {pipe=}, {self.current=}, subscribers=({n_subs}), children=({n_children}))"

    @property
    def current(self) -> typing.Union[OUT, Default]:
        """The current value."""
        return self._value

    @property
    def subscribers(self) -> list["Subscription[OUT]"]:
        """All `Subscription`s listening to this `Subject`."""
        return self._subscribers

    @property
    def has_subscribers(self) -> bool:
        """Whether any `Subject` listens to this `Subscription`."""
        return self._total_subscribers > 0

    def subscribe(self) -> "Subscription[OUT]":
        """Add a `Subscription` that will be notified on `update`."""

        subscription = Subscription[OUT](self._maxsize, self)
        self._on_subscribe(subscription)

        self.subscribers.append(subscription)

        return subscription

    def on_subscribe(self) -> None:
        """
        Emit an event when the first subscription is added.

        Override this method to start external listeners or resources
        when the first subscriber begins listening.
        """

    def _on_subscribe(self, subscription: Subscription) -> None:
        """Emit an event when `subscribe` is called."""
        if not self.has_subscribers:
            self.on_subscribe()
        self._total_subscribers += 1
        if self._parent is not None:
            self._parent._on_subscribe(subscription)

    def unsubscribe(self, subscriber: "Subscription[typing.Any]") -> None:
        """Remove a `Subscription`."""
        if subscriber in self.subscribers:
            subscriber.cancel()
            self.subscribers.remove(subscriber)
            self._on_unsubscribe()
            return

        for child in self._children:
            with contextlib.suppress(ValueError):
                child.unsubscribe(subscriber)
                return

        msg = "Subscription not found in subscribers or children."
        raise ValueError(msg)

    def on_unsubscribe(self) -> None:
        """
        Emit an event when the last subscription is removed.

        Override this method to perform cleanup or release resources when
        there are no active subscribers.
        """

    def _on_unsubscribe(self) -> None:
        """Emit an event when `unsubscribe` is called."""
        self._total_subscribers -= 1
        if not self.has_subscribers:
            self.on_unsubscribe()
        if self._parent is not None:
            if self._is_temporary and not self.subscribers and not self._children:
                self._parent._children.remove(self)
            self._parent._on_unsubscribe()

    def notify(self) -> None:
        """Propagate the current value to all listening `Subscription`s."""
        value = self._value
        if not isinstance(value, Default):
            value = typing.cast(OUT, value)
            for child in self._children:
                child.update(value)
            for subscriber in self.subscribers:
                subscriber.update(value)

    def update(self, value: IN) -> None:
        """Update the current value and `notify` all listening `Subscription`s."""

        self._value = self._pipe(value)
        self.notify()

    def pipe(self, func: typing.Callable[[OUT], T], temporary: bool = False) -> "Subject[OUT, T]":
        """
        Create a new `Subject` with `func` added to the list of pipes that are applied to values recieved from `notify`.

        Args:
          func: The callable that should be applied to all values seen by the new `Subject`.
          temporary: Whether or not the pipe should be pruned from its parent on `unsubscribe`.

        Returns:
          A new `Subject` with the pipe function added.

        Examples:
          Chain multiple pipe functions:
          >>> def first_pipe(x: str) -> str:
          ...     return x.upper()
          >>> def second_pipe(x: str) -> dict[str, str]:
          ...     return {"value": x}
          >>> subject = Subject[str]()
          >>> piped = subject.pipe(first_pipe).pipe(second_pipe)
          >>> async for value in piped.subscribe():
          ...     print(value)
          Here a `value` received from `piped.subscribe()` is equivalent to `second_pipe(first_pipe(x))`
          where `x` is the value received from `Subject.update`.

          Create multiple subjects with different pipes that are simultaneously updated:
          >>> subject = Subject[int]()
          >>> plus_one = subject.pipe(lambda x: x + 1)
          >>> times_two = subject.pipe(lambda x: x * 2)
          >>> subject.update(3)
          >>> await plus_one.get()  # 4
          >>> await times_two.get()  # 6
        """
        if not temporary and self._is_temporary:
            msg = (
                "Cannot create a non-temporary `Subject` from a temporary `Subject`, use pipe() with `temporary=True` "
                "or adjust the current `Subject` to be non-temporary."
            )
            raise RuntimeError(msg)

        new_subject = Subject[OUT, T](maxsize=self._maxsize, pipe=func)
        self._children.append(new_subject)
        new_subject._parent = weakref.proxy(self)
        new_subject._is_temporary = temporary
        return new_subject

    def filter(
        self, predicate: typing.Callable[[OUT], typing.Union[bool, typing.Any]], temporary: bool = False
    ) -> "Subject[OUT, OUT]":
        """
        Create a new `Subject` that is only notified when the item passes the `predicate`.

        Args:
          predicate: A filter predicate to apply.
          temporary: Whether the not the filter should be pruned from its parent on `unsubscribe`.

        Returns:
          A new `Subject` with the filter applied.

        Examples:
          Filter a subject
          >>> subject = Subject[int]()
          >>> filtered = subject.filter(lambda x: x > 5)
          >>> async for value in filtered.subscribe():
          ...     print(value)
          Here `filtered` only receives updates of numbers greater than 5.
        """

        def pipe(item: OUT) -> OUT:
            if bool(predicate(item)):
                return item

            return typing.cast(OUT, _DEFAULT_VALUE)

        return self.pipe(pipe, temporary=temporary)

    @typing.override
    def __enter__(self) -> "Subscription[OUT]":
        """
        Return a new `Subscription` upon entering the runtime context.

        Returns:
          The newly created `Subscription`.

        Examples:
          Subscribe to `my_subject`:
          >>> with my_subject as subscription:
          ...     async for value in subscription:
          ...         print(value)
          When leaving the runtime context, `subscription` is unsubscribed from `my_subject`.
        """
        if self._context is not None:
            msg = (
                "This Subject is already entered in a context. "
                "You cannot enter the same Subject multiple times simultaneously. "
                "Use the existing subscription or exit the previous context first."
            )
            raise RuntimeError(msg)

        self._context = self.subscribe()
        return self._context

    @typing.override
    def __exit__(
        self,
        exc_type: typing.Optional[type[BaseException]] = None,
        exc_value: typing.Optional[BaseException] = None,
        traceback: typing.Optional[types.TracebackType] = None,
    ) -> bool:
        if self._context:
            self.unsubscribe(self._context)
            self._context = None

        return False
